import sys
import time
import ctypes
from collections import deque
from fsgs.platform import PlatformHandler
from game_center.resources import resources
from fsbc.system import windows
from fsgs.util.gamenameutil import GameNameUtil
from game_center.gamecenter import GameCenter
from game_center.main import Main
from game_center.resources import logger
from game_center.glui.opengl import gl, fs_emu_texturing, fs_emu_blending
from game_center.glui.settings import Settings
from game_center.glui.sdl import SDL_IsMinimized
from game_center.glui.imageloader import ImageLoader
from game_center.glui.itemmenu import ItemMenu
from game_center.glui.displaylists import DisplayLists
from game_center.glui.bezier import Bezier
from game_center.glui.input import InputHandler
from game_center.glui.animation import AnimationSystem
from game_center.glui.animation import AnimateValueBezier
from game_center.glui.state import State
from game_center.glui.render import Render
from game_center.glui.notification import NotificationRender
from game_center.glui.font import Font, BitmapFont
from game_center.glui.items import MenuItem, AllMenuItem, NoItem, PlatformItem
from game_center.glui.texture import Texture
from game_center.glui.texturemanager import TextureManager
from game_center.glui.constants import TOP_HEIGHT

# FIXME: rename to manager or director or somethign
main_window = None

# ENABLE_VSYNC = Config.get_bool("Menu/VSync", True)
ENABLE_VSYNC = "--vsync" in sys.argv
ALWAYS_RENDER = True
# IDLE = Config.get_bool("Menu/Idle", 1)
IDLE = 1
# Render.display_fps = pyapp.user.ini.get_bool("Menu/ShowFPS", False)
# LIGHTING = Config.get_bool("Menu/Lighting", False)
LIGHTING = False
RENDER_DEBUG_SQUARES = 0
SEARCH_CHARS = \
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789 -:.,"
CONFIG_SEPARATION = 0.15
ANIMATION_SETTLE_TIME_1 = 1.0
ANIMATION_SETTLE_TIME_2 = 2.0


force_display_ratio = None
display = None
real_display_height = 0
current_menu = None
last_menu = None
last_game = None


def get_current_time():
    return time.time()


def set_current_menu(menu):
    if not isinstance(menu, SearchResultsMenu):
        menu.search_text = ""

    global current_menu
    if len(menu) == 0:
        menu.append(NoItem())
    current_menu = menu
    State.current_menu = menu
    if menu.navigatable:
        State.navigatable = menu.navigatable
    else:
        State.navigatable = menu
    State.top_menu = menu.top
    State.down_navigatable = None
    Render.dirty = True


# WINDOWED_SIZE = [1024, 640]  # 16:10
# WINDOWED_SIZE = (1024, 576)  # 16:9
# WINDOWED_SIZE = [1024, 768]  # 4:3
# WINDOWED_SIZE = [960, 540]  # 16:9
WINDOWED_SIZE = [1280, 720]  # 16:9


RENDER_GAME = ["IMAGE", "REFLECTIONS", "SCREENSHOTS"]
RENDER_GAME_OVERLAY = ["TITLE", "HEADER"]

# USE_MENU_TRANSITIONS = pyapp.user.ini.get_bool("Menu/Transitions", True)
# if "--disable-shaders" in sys.argv:
#     USE_MENU_TRANSITIONS = False
USE_MENU_TRANSITIONS = False
FIELD_COLOR = 0.1
WALL_COLOR = 0.0
# IDLE_EVENT = pygame.NUMEVENTS - 1
DEFAULT_ITEM_BRIGHTNESS = 0.8


class MenuGameTransition(object):
    value = 0.0


class MenuForwardTransition(object):
    # must be initially set to 1 because this indicates that
    # the transition is done / not in progress
    value = 1.0


class MenuBackwardTransition(object):
    # must be initially set to 1 because this indicates that
    # the transition is done / not in progress
    value = 1.0


class ScreenshotTransition(object):
    anim = None
    # must be initially set to 1 because this indicates that
    # the transition is done / not in progress
    value = 1.0


class RunTransition(object):
    value = 0.0


class Mouse(object):

    items = []
    focus = None

    @classmethod
    def set_visible(cls, visible=True):
        assert visible is not None
        # if State.mouse_visible == visible:
        #     return
        # if not visible:
        #     cls.focus = None
        # pygame.mouse.set_visible(visible)
        # State.mouse_visible = visible
        print("Mouse.set_visible not implemented")


def set_items_brightness(brightness, duration=1.0, delay=0.0):
    State.items_brightness_anim = AnimateValueBezier(
        (State, "items_brightness"),
        State.items_brightness, State.time + delay,
        State.items_brightness, State.time + delay,
        brightness, State.time + delay + duration,
        brightness, State.time + delay + duration)


def compile_shader(source, shader_type):
    shader = gl.glCreateShaderObjectARB(shader_type)
    gl.glShaderSourceARB(shader, source)
    gl.glCompileShaderARB(shader)
    try:
        status = ctypes.c_int()
        gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS, ctypes.byref(status))
        status = status.value
    except TypeError:
        status = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)
    if not status:
        print_log(shader)
        gl.glDeleteObjectARB(shader)
        raise ValueError("Shader compilation failed")
    return shader


def compile_program(vertex_source, fragment_source):
    vertex_shader = None
    fragment_shader = None
    program = gl.glCreateProgram()
    if vertex_source:
        print("compile vertex shader")
        vertex_shader = compile_shader(vertex_source, gl.GL_VERTEX_SHADER)
        gl.glAttachShader(program, vertex_shader)
    if fragment_source:
        print("compile fragment shader")
        fragment_shader = compile_shader(fragment_source, gl.GL_FRAGMENT_SHADER)
        gl.glAttachShader(program, fragment_shader)
    gl.glLinkProgram(program)
    status = gl.glGetProgramiv(program, gl.GL_LINK_STATUS)
    if status == gl.GL_FALSE:
        print("could not link shader program")
        print(gl.glGetProgramInfoLog(program))
        sys.exit(1)
    if vertex_shader:
        gl.glDeleteShader(vertex_shader)
    if fragment_shader:
        gl.glDeleteShader(fragment_shader)
    return program


def print_log(shader):
    length = ctypes.c_int()
    gl.glGetShaderiv(shader, gl.gl.GL_INFO_LOG_LENGTH, ctypes.byref(length))
    if length.value > 0:
        shader_log = gl.glGetShaderInfoLog(shader)
        print(shader_log)
        sys.exit(1)


texture_program = None
color_program = None
premultiplied_texture_program = None


def compile_programs():
    global texture_program, color_program, premultiplied_texture_program
    vertex_shader = None

    color_program = compile_program(vertex_shader, ["""
void main()
{
vec4 s = gl_Color;
float extra_alpha = 1.0;
//float extra_alpha = gl_TextureEnvColor[0].a;
float a = s.a * extra_alpha;
gl_FragColor.a = a;
gl_FragColor.r = s.r*a;
gl_FragColor.g = s.g*a;
gl_FragColor.b = s.b*a;
}
            """])

    texture_program = compile_program(vertex_shader, ["""
uniform sampler2D texture;

void main()
{
vec4 s = texture2D(texture,gl_TexCoord[0].st);
//float extra_alpha = 1.0;
//float extra_alpha = gl_TextureEnvColor[0].a;
float opacity = gl_Color.a;
float a = s.a * opacity;
gl_FragColor.a = a;
gl_FragColor.r = s.r * a;
gl_FragColor.g = s.g * a;
gl_FragColor.b = s.b * a;
}
"""])

    premultiplied_texture_program = compile_program(vertex_shader, ["""
uniform sampler2D texture;

void main()
{
vec4 s = texture2D(texture,gl_TexCoord[0].st);
//float extra_alpha = 1.0;
//float extra_alpha = gl_TextureEnvColor[0].a;
float opacity = gl_Color.a;
gl_FragColor.a = s.a * opacity;
gl_FragColor.r = s.r * opacity;
gl_FragColor.g = s.g * opacity;
gl_FragColor.b = s.b * opacity;
}
"""])


def enable_texture_shader():
    fs_emu_blending(True)
    gl.glBlendFunc(gl.GL_ONE, gl.GL_ONE_MINUS_SRC_ALPHA)
    gl.glUseProgram(texture_program)


def disable_shader():
    gl.glUseProgram(0)


def set_program(program):
    if not program:
        gl.glUseProgram(0)
    else:
        gl.glUseProgram(program)


def enter_menu(result, replace=False):
    print("enter_menu", result, "replace", replace)
    print("   menu parent_menu", result.parent_menu)

    if replace:
        State.history.pop()
    c_menu = State.history[-1]
    result.parents[:] = c_menu.parents
    result.parents.append(c_menu)
    print("   menu parents    ", result.parents)
    State.history.append(result)
    set_current_menu(result)


def create_main_menu():
    new_menu = AllMenuItem().activate(None)
    if len(new_menu) == 0:
        new_menu.append(NoItem())
    return new_menu


def recreate_main_menu_if_necessary():
    new_menu = create_main_menu()
    State.history.append(new_menu)
    set_current_menu(new_menu)


def show():
    global current_menu

    # fade_from is used on init_display, so we initialize this
    # color here. Set alpha to 2.0 to force 1 second of solid
    # color in combination with 2 sec. animation below
    if False and windows and not Settings.fullscreen_menu:
        State.fade_from = (1.0, 1.0, 1.0, 2.0)
        State.fade_to = (1.0, 1.0, 1.0, 0.0)
    else:
        State.fade_from = (0.0, 0.0, 0.0, 2.0)
        State.fade_to = (0.0, 0.0, 0.0, 0.0)
    init_display()
    if LIGHTING:
        init_lighting()
    init_textures()
    init_fonts()

    if USE_MENU_TRANSITIONS:
        compile_programs()

    InputHandler.open()

    on_resize((Render.display_width, Render.display_height))
    image_loader = ImageLoader.get()
    image_loader.start()
    new_menu = create_main_menu()
    State.history.append(new_menu)

    for platform_id in PlatformHandler.get_platform_ids():
        if "--" + platform_id in sys.argv:
            platform_menu = ItemMenu()
            platform_menu.parent_menu = new_menu

            platform_item = PlatformItem(platform_id)
            platform_menu.items.append(platform_item)
            # platform_menu.set_selected_index(0, immediate=True)

            new_menu = platform_item.activate(platform_menu)
            print(new_menu)
            State.history.append(new_menu)
            break

    set_current_menu(new_menu)
    if len(new_menu) == 1:
        # only n/a item showing, possibly
        if len(AllMenuItem().activate(None)) == 0:
            # no games, initiate game scan
            rescan_games()

    State.fade_start = get_current_time()
    State.fade_end = get_current_time() + 2.000

    # # make a timer so that update events are sent to modules at least once
    # # every second
    # pygame.time.set_timer(IDLE_EVENT, 1000)

    State.start_time = get_current_time()


def init_fonts():
    NotificationRender.init()

    BitmapFont.title_font = BitmapFont("title_font")
    BitmapFont.menu_font = BitmapFont("menu_font")


liberation_sans_bold_path = None
liberation_sans_narrow_bold_path = None
vera_font_path = None


def reinit_fonts():
    global liberation_sans_bold_path
    global liberation_sans_narrow_bold_path
    global vera_font_path

    if liberation_sans_bold_path is None:
        liberation_sans_bold_path = resources.resource_filename(
            "LiberationSans-Bold.ttf")
    if liberation_sans_narrow_bold_path is None:
        liberation_sans_narrow_bold_path = resources.resource_filename(
            "LiberationSansNarrow-Bold.ttf")
    if vera_font_path is None:
        vera_font_path = resources.resource_filename(
            "VeraBd.ttf")

    if Font.title_font is None:
        Font.title_font = Font(
            liberation_sans_bold_path, int(0.04 * Render.display_height))
    if Font.subtitle_font is None:
        Font.subtitle_font = Font(
            liberation_sans_bold_path, int(0.025 * Render.display_height))
    if Font.small_font is None:
        Font.small_font = Font(
            liberation_sans_narrow_bold_path, int(0.025 * Render.display_height))
    if Font.main_path_font is None:
        Font.main_path_font = Font(
            liberation_sans_bold_path, int(0.025 * Render.display_height))
    if Font.list_subtitle_font is None:
        Font.list_subtitle_font = Font(
            liberation_sans_bold_path, int(0.020 * Render.display_height))
    if Font.header_font is None:
        Font.header_font = Font(
            vera_font_path, int(0.06 * Render.display_height))

    Font.title_font.set_size(int(0.04 * Render.display_height))
    Font.subtitle_font.set_size(int(0.025 * Render.display_height))
    Font.small_font.set_size(int(0.025 * Render.display_height))
    Font.main_path_font.set_size(int(0.025 * Render.display_height))
    Font.list_subtitle_font.set_size(int(0.020 * Render.display_height))
    Font.header_font.set_size(int(0.06 * Render.display_height))


char_buffer = ""
char_buffer_last_updated = 0


def handle_search_menu_on_character_press(char_buffer):
    # if len(char_buffer) > 2 and t - char_buffer_last_updated > 0.5:
    if len(char_buffer) > 2:
        print(char_buffer)
        create_search_results_menu(char_buffer)
    if len(char_buffer) <= 2:
        if isinstance(current_menu, SearchResultsMenu):
            # collapse search menu

            # if hasattr(current_menu, "search_text"):
            print("setting current_menu (something with search)")
            # set_current_menu(current_menu.parent_menu)
            go_back()
        else:
            current_menu.search_text = char_buffer


# noinspection PyUnusedLocal
def character_press(char):
    print("character press", repr(char))
    if State.current_game:  # or isinstance(current_menu, GameMenu):
        print("ignoring key press", repr(char))
        return

    # global char_buffer
    char_buffer = current_menu.search_text

    global char_buffer_last_updated
    char_buffer_last_updated = State.time
    # return

    Render.dirty = True
    if char == "RETURN":
        print(char_buffer, len(char_buffer))
        print("returning false")
        return False
    elif char == "BACKSPACE" or char == u"\b":
        char_buffer = char_buffer[:-1]
        if len(char_buffer) <= 2:
            char_buffer = ""
        handle_search_menu_on_character_press(char_buffer)
        return True
    elif len(char) == 1:
        char_buffer += char
        handle_search_menu_on_character_press(char_buffer)
        if 1 <= len(char_buffer) <= 2:
            jump_to_item = -1
            for i, item in enumerate(current_menu.items):
                # FIXME: a bit hack-y this, should check sort_name
                # instead (need to store this in items then)
                # also, should use binary search and not sequential search...
                searches = [char_buffer, "the " + char_buffer,
                            "a " + char_buffer]
                check = item.name.lower()
                for s in searches:
                    if check.startswith(s):
                        jump_to_item = i
                        break
            if jump_to_item >= 0:
                current_menu.set_selected_index(jump_to_item, immediate=True)

        return True
    else:
        raise Exception(repr(char))


class SearchTextItem(MenuItem):

    def __init__(self, title):
        MenuItem.__init__(self)
        self.path_title = title


class SearchResultsMenu(ItemMenu):
    pass


def create_search_results_menu(text):
    global current_menu
    try:
        if text == current_menu.search_text:
            return False
    except AttributeError:
        pass
    new_menu = SearchResultsMenu("Search Results")
    new_menu.search_text = text
    # words = [v.strip() for v in text.lower().split(" ")]
    # print "Creating search results for", words
    new_menu.top.append_left(
        SearchTextItem("Search: {0}_".format(text)))
    # clause = []
    # args = []
    # for word in words:
    #     clause.append("AND name like ?")
    #     args.append("%{0}%".format(word))
    # clause = " ".join(clause)
    terms = GameNameUtil.extract_search_terms(text.lower())
    for item in MenuItem.create_game_items(terms):
        new_menu.append(item)
    if len(new_menu) == 0:
        new_menu.append(NoItem("No Search Results"))
    # if hasattr(current_menu, "search_text"):
    #     # replace current search menu, not append to path
    #     #new_menu.parent_menu = current_menu.parent_menu
    #     replace = True
    # else:
    #     #new_menu.parent_menu = current_menu
    #     replace = False
    replace = isinstance(current_menu, SearchResultsMenu)
    print("create search results menu")
    # set_current_menu(new_menu)
    enter_menu(new_menu, replace=replace)
    return True


def rescan_games():
    # global current_menu
    print("rescan games -- currently disabled")
    # GameScanner.scan()
    # Render.dirty = True
    pass


def go_back():
    c_menu = State.history[-1]
    if hasattr(c_menu, "go_back"):
        c_menu.go_back()
        return
    State.history.pop()
    set_current_menu(State.history[-1])


def default_input_func(button):
    if button == "LEFT":
        State.navigatable.go_left()
    elif button == "RIGHT":
        State.navigatable.go_right()
    elif button == "UP":
        State.navigatable.go_up()
    elif button == "DOWN":
        State.navigatable.go_down()
    elif button == "SKIP_LEFT":
        State.navigatable.go_left(10)
    elif button == "SKIP_RIGHT":
        State.navigatable.go_right(10)
    elif button == "PRIMARY":

        # global char_buffer
        # char_buffer = ""

        State.navigatable.activate()

    elif button == "BACK":
        print("-- button is BACK -- ")
        if State.config_menu:
            State.config_menu = None
            set_current_menu(State.current_menu)
            set_items_brightness(0.66, duration=0.500)
        elif State.current_game is None and current_menu.search_text:
            InputHandler.get_button()  # clear OK status
            character_press("BACKSPACE")
        elif State.current_game:
            State.current_game = None
            # game_fade_animation = AnimateValueBezier(
            #     (MenuGameTransition, "value"),
            #     1.0, State.time,
            #     1.0, State.time + 0.133,
            #     0.0, State.time + 0.133,
            #     0.0, State.time + 0.400)

        # elif can_navigate and current_menu.parent_menu:
        # elif can_navigate and len(State.history) > 1:
        elif len(State.history) > 1:
            go_back()
    elif button == "QUIT":
        State.quit = True


def render_top():
    Render.hd_perspective()
    gl.glPushMatrix()
    transition = State.current_menu.top_menu_transition
    gl.glTranslate(0.0, (1.0 - transition) * 90, 0.9)

    gl.glTranslate(0.0, 0.0, 0.05)

    if State.top_menu == State.navigatable:
        selected_index = State.top_menu.get_selected_index()
    else:
        selected_index = -1
    x = State.gl_left
    for item in State.top_menu.left:
        item.update_size_left()
        item.x = x
        item.y = 1080 - TOP_HEIGHT
        item.h = TOP_HEIGHT
        x += item.w

    index = len(State.top_menu.left) - 1
    for item in reversed(State.top_menu.left):
        item.render_top_left(selected=(index == selected_index))
        index -= 1

    index = len(State.top_menu) - 1
    x = State.gl_right
    for item in reversed(State.top_menu.right):
        item.update_size_right()
        x -= item.w
        item.x = x 
        item.y = 1080 - TOP_HEIGHT
        item.h = TOP_HEIGHT
        if Mouse.focus:
            selected = (Mouse.focus == item)
        else:
            selected = (index == selected_index)
        item.render_top_right(selected=selected)
        Mouse.items.append(item)
        index -= 1
    gl.glPopMatrix()


def render_config_menu():
    if not State.config_menu:
        return

    Render.ortho_perspective()
    config_menu = State.config_menu
    # text = config_menu.items[config_menu.index]
    # otw, th = Render.text(text, title_font,
    #         -1.0, -0.93, 2.0, color=(1.0, 1.0, 1.0, 0.36 * strength))
    # x = 0.0 + otw / 2 + CONFIG_SEPARATION
    # for i in range(config_menu.index + 1, len(config_menu.items)):
    #     text = config_menu.items[i]
    #     tw, th = Render.text(text, title_font,
    #             x, -0.93, color=(1.0, 1.0, 1.0, 0.36 * strength))
    #     x += tw + CONFIG_SEPARATION
    # x = 0.0 - otw / 2 - CONFIG_SEPARATION
    x = -0.55
    y = 0.8
    for i in range(len(config_menu.items)):
        text = config_menu.items[i].upper()
        # tw, th = Render.measure_text(text, title_font)
        # x -= tw + CONFIG_SEPARATION
        y -= 0.15
        if i == config_menu.index and config_menu == State.navigatable:
            color = (1.0, 1.0, 1.0, 1.0)
        else:
            color = (1.0, 1.0, 1.0, 0.33)
        Render.text(text, Font.subtitle_font, x, y, color=color)


def render_scanning_status():
    # Render.hd_perspective()
    # text = GameScanner.get_status()
    # Render.dirty = True
    #
    # fs_emu_texturing(False)
    #
    # z = 0.0
    #
    # glBegin(GL_QUADS)
    # glColor3f(0.0, 0.0, 0.0)
    # glVertex3f(   0, 500, z)
    # glVertex3f(1920, 500, z)
    # glVertex3f(1920, 700, z)
    # glVertex3f(   0, 700, z)
    # glEnd()
    #
    # Render.text(text, Font.title_font, 200, 600, color=(1.0, 1.0, 1.0, 1.0))
    #
    # glBegin(GL_QUADS)
    # glColor3f(1.0, 1.0, 1.0)
    # x = 200
    # y = 500
    # z = 0.9
    # x2 = 200 + 1520 * GameScanner.progress
    # glVertex3f(x, y, z)
    # glVertex3f(x2, y, z)
    # glVertex3f(x2, y + 20, z)
    # glVertex3f(x, y + 20, z)
    # glEnd()
    # fs_emu_texturing(True)
    pass


def do_render():
    if current_menu is None:
        return

    current_menu.update()

    if RunTransition.value > 0.99:
        # do not render anything when running a game
        return

    if State.currently_ingame:
        # print("currently ingame")
        return

    # try to exploit parallelism by uploading texture while rendering
    TextureManager.get().load_textures(1)

    # clear mouseover rects -these will be calculated during rendering
    Mouse.items[:] = []
    Render.standard_perspective()

    # scanning = GameScanner.scanning

    data = current_menu.render()
    current_menu.render_transparent(data)

    # if GameScanner.is_scanning():
    if False:
        render_scanning_status()
        State.was_scanning = True
    else:
        render_config_menu()
        if State.was_scanning:
            print("State.was_scanning")
            State.was_scanning = False
            # reload current menu

            # if current_menu.parent_menu:
            #     result = current_menu.parent_menu.selected_item.activate(
            #             current_menu.parent_menu)
            #     if isinstance(result, Menu):
            #         #if len(result) == 0:
            #         #    result.append(NoItem())
            #         result.parent_menu = current_menu.parent_menu
            #         print("set new menu (rescanned games)")
            #         set_current_menu(result)
            recreate_main_menu_if_necessary()

    render_top()

    if State.dialog:
        State.dialog.render()

    render_global_fade()
    NotificationRender.render()
    if RunTransition.value > 0.0:
        render_fade(a=RunTransition.value)


def render_fade(r=0.0, g=0.0, b=0.0, a=0.0):
    Render.hd_perspective()
    fs_emu_blending(True)
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor4f(r * a, g * a, b * a, a)
    gl.glVertex2f(0, 0)
    gl.glVertex2f(1920, 0)
    gl.glVertex2f(1920, 1080)
    gl.glVertex2f(0, 1080)
    gl.glEnd()


def render_global_fade():
    t = State.time
    if State.fade_end >= t >= State.fade_start:
        a = (t - State.fade_start) / (State.fade_end - State.fade_start)
        if a < 0.0:
            a = 0.0
        elif a > 1.0:
            a = 1.0

        Render.hd_perspective()
        gl.glPushMatrix()
        gl.glTranslatef(0.0, 0.0, 0.99999)
        fs_emu_blending(True)
        fs_emu_texturing(True)
        if State.fade_splash:
            Texture.splash.render(
                (1920 - Texture.splash.w) // 2,
                (1080 - Texture.splash.h) // 2, Texture.splash.w,
                Texture.splash.h, opacity=(1.0 - a))
        
        c = [0, 0, 0, (1.0 - a)]
        # for i in range(4):
        #     c[i] = State.fade_from[i] + \
        #             (State.fade_to[i] - State.fade_from[i]) * a * (a)
        render_fade(*c)
        gl.glPopMatrix()
        Render.dirty = True


FPS_FRAMES = 100
render_times = deque()
for _ in range(FPS_FRAMES):
    render_times.append(0)

fps_str = ""
debug_x = 0
debug_x_2 = 0


def render_debug_square():
    global debug_x
    Render.hd_perspective()
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(1.0, 1.0, 1.0)
    x = debug_x
    debug_x += 1
    if debug_x >= 1920:
        debug_x = 0
    y = 989
    z = 0.99
    gl.glVertex3f(x, y, z)
    gl.glVertex3f(x + 20, y, z)
    gl.glVertex3f(x + 20, y + 20, z)
    gl.glVertex3f(x, y + 20, z)
    gl.glEnd()
    fs_emu_texturing(True)


def render_debug_square_2():
    global debug_x_2
    Render.hd_perspective()
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(0.2, 0.2, 0.2)
    x = debug_x_2
    debug_x_2 += 1
    if debug_x_2 >= 1920:
        debug_x_2 = 0
    y = 989 + 5
    z = 0.99
    gl.glVertex3f(x, y, z)
    gl.glVertex3f(x + 20, y, z)
    gl.glVertex3f(x + 20, y + 10, z)
    gl.glVertex3f(x, y + 10, z)
    gl.glEnd()
    fs_emu_texturing(True)


# FIXME: remove / move some code away
def swap_buffers():
    # global fps_str
    #
    # Render.ortho_perspective()
    # #glPushMatrix()
    # #glTranslatef(0.0, 0.0, 0.5)
    #
    # #if not fps_str:
    # #    fps_str = "WHY?"
    #
    # if Render.display_fps or True:
    #     if fps_str:
    #         Render.text(fps_str, Font.main_path_font,
    #                     -1.74, 0.82, h=0.1, color=(0.25, 0.25, 0.25, 1.0))
    # #glPopMatrix()
    #
    # # FIXME: Why does not minimize from fullscreen work on ATI unless we render
    # # something here?
    # glBindTexture(GL_TEXTURE_2D, 0)
    # glPushMatrix()
    # glTranslate(2000.0, 0.0, 0.0)
    # glBegin(GL_QUADS)
    # glVertex2f(0.0, 0.0)
    # glVertex2f(1.0, 0.0)
    # glVertex2f(1.0, 1.0)
    # glVertex2f(0.0, 1.0)
    # glEnd()
    # glPopMatrix()
    #
    # #fs_emu_blending(False)
    # #glEnable(GL_DEPTH_TEST)
    #
    # #if Render.display_sync:
    # #    glFinish()
    #
    # #pygame.display.flip()
    # print("FIXME: not flipping")
    #
    # if Render.display_sync:
    #     # give up time slice
    #     time.sleep(0.001)
    #     glFinish()
    #
    # if Render.display_fps:
    #     t = get_current_time()
    #     render_times.append(t)
    #     t0 = render_times.popleft()
    #     if t0 > 0 and State.frame_number % 5 == 0:
    #         time_diff = t - t0
    #         #print("{0:0.2f}".format(300.0 / time_diff))
    #         fps = FPS_FRAMES / time_diff
    #         if fps >= 100:
    #             fps_str = "FPS:  {0:0.0f}".format(fps)
    #         else:
    #             fps_str = "FPS: {0:0.1f}".format(fps)
    #         #Render.text(fps_str, Font.title_font,
    #         #        -1.0, 0.90, 2.0, shadow=True)
    State.frame_number += 1
    Render.delete_textures()


def render_screen():
    # set Render.dirty to False here, so that render functions can
    # request a new render frame by setting dirty
    Render.dirty = False
    # glEnable(GL_SCISSOR_TEST)
    # can_idle =
    if SDL_IsMinimized():
        time.sleep(0.01)
        return
    # Render.dirty = True
    gl.glClearColor(0.0, 0.0, 0.0, 1.0)
    gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
    do_render()
    if Render.display_fps:
        Render.dirty = True
    if RENDER_DEBUG_SQUARES:
        render_debug_square()


# noinspection PyPep8Naming
def init_display():
    global display
    global real_display_height  # , display_yoffset

    # global banner_texture, shadow_texture, gloss_texture
    # global top_texture, top_logo_texture, logo_texture
    # global missing_cover_texture, default_item_texture
    # global backdrop_texture

    logger.debug("Init OpenGL menu display")

    DisplayLists.clear()

    # on_resize()
    # depth = 0
    # FIXME: HACK / TESTING
    # if not Settings.fullscreen_menu:
    #    if windows:
    #        os.environ["SDL_VIDEO_WINDOW_POS"] = "3,29"
    #    else:
    #        os.environ["SDL_VIDEO_WINDOW_POS"] = "0,0"
    # maximize_window = (not Settings.fullscreen_menu and
    #                    Settings.windowed_size is None)

    # display_info = pygame.display.Info()
    # dw = display_info.current_w
    # dh = display_info.current_h
    # dw, dh = fsui.get_screen_size()
    # dw, dh = 100, 100
    Render.display_width = main_window.width
    Render.display_height = main_window.height

    # if Settings.fullscreen_menu:
    #     print("fullscreen is True")
    #     if windows:
    #         #resolution = (0, 0)
    #         #flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME \
    #         #        | pygame.FULLSCREEN
    #         os.environ["SDL_VIDEO_WINDOW_POS"] = "0,0"
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME
    #         #flags = flags | pygame.FULLSCREEN
    #         #if fs.linux:
    #         #    pass
    #         #else:
    #         if dw > dh * 2:
    #             # Assume dual monitor setup - hack for Linux / SDL
    #             resolution = (dw / 2, dh)
    #         else:
    #             resolution = (dw, dh)
    #     else:  # fullscreen, but not microsoft windows
    #         #resolution = (0, 0)
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME #| pygame.FULLSCREEN
    #         if dw > dh * 2:
    #             # Assume dual monitor setup - hack for Linux / SDL
    #             resolution = (dw / 2, dh)
    #         else:
    #             resolution = (dw, dh)
    #         if linux:
    #             overscan = Config.get("display/overscan", "0,0,0,0")
    #             try:
    #                 overscan = overscan.split(",")
    #                 overscan = [int(x.strip()) for x in overscan]
    #                 print("using overscan", overscan)
    #             except Exception as e:
    #                 print("error parsing overscan from config:", repr(e))
    #                 overscan = [0, 0, 0, 0]
    #             os.environ["SDL_VIDEO_WINDOW_POS"] = "{0},{1}".format(
    #                 overscan[0], overscan[1])
    #             resolution = (resolution[0] - overscan[0] - overscan[2],
    #                           resolution[1] - overscan[1] - overscan[3])
    #         elif macosx:
    #             # FIXME: fullscreen mode does not work very well. -When
    #             # opening a fullscreen emulator from fullscreen, the emulator
    #             # crashes on glViewport. Tested with fs-amiga.
    #             #flags |= pygame.FULLSCREEN
    #
    #             # for now, we create an almost maximized window, works
    #             # quite well when the dock is set to auto-hide
    #             #resolution = (resolution[0], resolution[1] - 22)
    #
    #             # FIXME: trying LSUIPresentationMode
    #             os.environ["SDL_VIDEO_WINDOW_POS"] = "0,0"
    #
    #             # kUIModeNormal = 0
    #             # kUIModeContentSuppressed = 1
    #             # kUIModeContentHidden = 2
    #             # kUIModeAllSuppressed = 4
    #             kUIModeAllHidden = 3
    #             kUIOptionAutoShowMenuBar = 1 << 0
    #             # kUIOptionDisableAppleMenu = 1 << 2
    #             # kUIOptionDisableProcessSwitch = 1 << 3
    #             # kUIOptionDisableForceQuit = 1 << 4
    #             # kUIOptionDisableSessionTerminate = 1 << 5
    #             # kUIOptionDisableHide = 1 << 6
    #
    #             #noinspection PyUnresolvedReferences
    #             import objc
    #             #noinspection PyUnresolvedReferences
    #             from Foundation import NSBundle
    #             bundle = NSBundle.bundleWithPath_(
    #                 "/System/Library/Frameworks/Carbon.framework")
    #             objc.loadBundleFunctions(
    #                 bundle, globals(),
    #                 ((str("SetSystemUIMode"), str("III"), str("")),))
    #             #noinspection PyUnresolvedReferences
    #             SetSystemUIMode(kUIModeAllHidden, kUIOptionAutoShowMenuBar)
    #
    # else:
    #     if windows and maximize_window and \
    #             not Settings.window_decorations:
    #         import ctypes
    #         SPI_GETWORKAREA = 48
    #
    #         class RECT(ctypes.Structure):
    #             _fields_ = [
    #                 ("left", ctypes.c_ulong),
    #                 ("top", ctypes.c_ulong),
    #                 ("right", ctypes.c_ulong),
    #                 ("bottom", ctypes.c_ulong)]
    #
    #         m = ctypes.windll.user32
    #         r = RECT()
    #         m.SystemParametersInfoA(SPI_GETWORKAREA, 0, ctypes.byref(r), 0)
    #         x = int(r.left)
    #         y = int(r.top)
    #         w = int(r.right) - int(r.left)
    #         h = int(r.bottom) - int(r.top)
    #         print(x, y, w, h)
    #         WINDOWED_SIZE[0] = w
    #         WINDOWED_SIZE[1] = h
    #         os.environ["SDL_VIDEO_WINDOW_POS"] = "{0},{1}".format(x, y)
    #         State.allow_minimize = False
    #
    #     if Settings.windowed_size:
    #         print("Settings.windowed_size", Settings.windowed_size)
    #         WINDOWED_SIZE[0] = Settings.windowed_size[0]
    #         WINDOWED_SIZE[1] = Settings.windowed_size[1]
    #         Render.display_width = WINDOWED_SIZE[0]
    #         Render.display_height = WINDOWED_SIZE[1]
    #         #if dw > 1400:
    #         #    Render.display_width = 1280
    #         #    Render.display_height = 720
    #     else:
    #         Render.display_width = WINDOWED_SIZE[0]
    #         Render.display_height = WINDOWED_SIZE[1]
    #     resolution = (Render.display_width, Render.display_height)
    #     #print(resolution)
    #     #sys.exit(1)
    #     if Settings.window_decorations:
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.RESIZABLE
    #     else:
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME
    #
    # display_yoffset = 0
    # Mouse.set_visible(False)
    #
    # pygame.display.gl_set_attribute(pygame.GL_STENCIL_SIZE, 8)
    # pygame.display.gl_set_attribute(pygame.GL_DEPTH_SIZE, 16)
    #
    # Render.display_sync = ENABLE_VSYNC
    # if Render.display_sync:
    #     print("enabling vertical sync")
    #     os.environ["__GL_SYNC_TO_VBLANK"] = "1"
    #     pygame.display.gl_set_attribute(pygame.GL_SWAP_CONTROL, 1)
    # else:
    #     os.environ["__GL_SYNC_TO_VBLANK"] = "0"
    #     pygame.display.gl_set_attribute(pygame.GL_SWAP_CONTROL, 0)
    # pygame.display.gl_set_attribute(pygame.GL_DOUBLEBUFFER, 1)
    # fsaa = Config.get_int("video/fsaa", 0)
    # if fsaa:
    #     pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS, 1)
    #     pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLESAMPLES, fsaa)
    # print("pygame set display mode", resolution, flags, depth)
    # display = pygame.display.set_mode(resolution, flags, depth)
    # if not Settings.fullscreen_game:
    #     try:
    #         del os.environ["SDL_VIDEO_WINDOW_POS"]
    #     except KeyError:
    #         pass
    #
    # if app.name == "fs-uae-arcade":
    #     pygame.display.set_caption("FS-UAE Arcade")
    # else:
    #     pygame.display.set_caption("FS Game Center")
    #
    # # FIXME: DISABLING MAXIMIZE FOR DEBUGGING
    # #maximize_window = False
    # if maximize_window:
    #     print("maximizing window")
    #     SDL_Maximize()
    #     for event in pygame.event.get():
    #         if event.type == pygame.VIDEORESIZE:
    #             #WINDOWED_SIZE[0] = event.w
    #             #WINDOWED_SIZE[1] = event.h
    #             on_resize((event.w, event.h))
    #     print("DISPLAY.GET_SIZE", display.get_size())
    # else:
    #     on_resize(display.get_size())

    gl.glMatrixMode(gl.GL_MODELVIEW)

    gl.glBlendFunc(gl.GL_ONE, gl.GL_ONE_MINUS_SRC_ALPHA)
    gl.glClearColor(*State.fade_from)

    fs_emu_texturing(True)
    Texture.splash = Texture("splash.png")

    for _ in range(2):
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        Render.hd_perspective()
        Texture.splash.render(
            (1920 - Texture.splash.w) // 2,
            (1080 - Texture.splash.h) // 2, Texture.splash.w,
            Texture.splash.h)
        gl.glFinish()
        # pygame.display.flip()
        gl.glFinish()

    gl.glEnable(gl.GL_DEPTH_TEST)


def init_textures():
    Texture.shadow = Texture.from_resource("shadow.png")
    Texture.shadow2 = Texture.from_resource("shadow2.png")
    # Texture.gloss = Texture.from_resource("gloss.png")
    Texture.gloss = Texture("gloss.png")
    Texture.screen_gloss = Texture("screen_gloss.png")
    Texture.static = Texture("preview_static0.png")
    Texture.default_item = Texture("default_item.png")
    Texture.missing_screenshot = Texture("missing_screenshot.png")
    Texture.missing_cover = Texture("missing_cover.png")
    # path = os.path.join(fs.get_data_dir(), "logo.png")
    # if os.path.exists(path):
    #     im = Image.open(path)
    #     Texture.logo = Texture.load(im)
    # else:
    Texture.logo = Texture.from_resource("logo.png")
    Texture.top = Texture.from_resource("top.png")
    Texture.top_logo = Texture("top_logo.png")
    Texture.top_logo_selected = Texture("top_logo_selected.png")

    Texture.add = Texture("add.png")
    Texture.add_selected = Texture("add_selected.png")
    Texture.home = Texture("home.png")
    Texture.home_selected = Texture("home_selected.png")
    Texture.minimize = Texture("minimize.png")
    Texture.minimize_selected = Texture("minimize_selected.png")
    Texture.close = Texture("close.png")
    Texture.close_selected = Texture("close_selected.png")
    Texture.shutdown = Texture("shutdown.png")
    Texture.shutdown_selected = Texture("shutdown_selected.png")

    Texture.bottom_bar = Texture("bottom_bar.png")
    Texture.screen_border_1 = Texture("screen_border_1.png")
    Texture.screen_border_2 = Texture("screen_border_2.png")
    Texture.top_background = Texture("top_background.png")
    Texture.top_item = Texture("top_item.png")
    Texture.top_item_selected = Texture("top_item_selected.png")
    Texture.top_item_left = Texture("top_item_left.png")
    Texture.top_item_left_selected = Texture("top_item_left_selected.png")
    Texture.top_item_right = Texture("top_item_right.png")
    Texture.top_item_arrow = Texture("top_item_arrow.png")
    Texture.top_item_arrow_selected = Texture("top_item_arrow_selected.png")
    
    Texture.sidebar_background_shadow = Texture(
        "sidebar_background_shadow.png")
    Texture.glow_top = Texture("glow_top.png")
    Texture.glow_top_left = Texture("glow_top_left.png")
    Texture.glow_left = Texture("glow_left.png")

    # # FIXME: TEMPORARY - FOR TESTING, ONLY
    # path = "c:\\git\\fs-game-database\\Backdrops\\ffx.png"
    # if os.path.exists(path):
    #     im = Image.open(path)
    #     #im = resources.get_resource_image_pil("shadow.png")
    #     assert im.size == (1024, 1024)
    #     imdata = im.tostring("raw", "RGB")
    #     backdrop_texture = Render.create_texture()
    #     glBindTexture(GL_TEXTURE_2D, backdrop_texture)
    #     glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, im.size[0], im.size[1], 0,
    #                  GL_RGB, GL_UNSIGNED_BYTE, imdata)
    #     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
    #     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    # else:
    #     backdrop_texture = 0


def init_lighting():
    gl.glLightModeli(gl.GL_LIGHT_MODEL_LOCAL_VIEWER, gl.GL_TRUE)
    gl.glLightModeli(gl.GL_LIGHT_MODEL_TWO_SIDE, gl.GL_FALSE)
    gl.glLightModeli(
        gl.GL_LIGHT_MODEL_COLOR_CONTROL, gl.GL_SEPARATE_SPECULAR_COLOR)

    light_position = (0.0, 0.0, 3.0, 1.0)
    gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, light_position)
    gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, (1.0, 1.0, 1.0, 1.0))
    gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, (0.0, 0.0, 0.0, 1.0))
    gl.glEnable(gl.GL_LIGHT0)

    gl.glLightfv(gl.GL_LIGHT1, gl.GL_DIFFUSE, (0.0, 0.0, 0.0, 1.0))
    gl.glLightfv(gl.GL_LIGHT1, gl.GL_SPECULAR, (1.0, 1.0, 1.0, 1.0))
    gl.glEnable(gl.GL_LIGHT1)

    gl.glLightfv(gl.GL_LIGHT2, gl.GL_DIFFUSE, (0.0, 0.0, 0.0, 1.0))
    gl.glLightfv(gl.GL_LIGHT2, gl.GL_SPECULAR, (0.5, 0.5, 0.5, 1.0))
    gl.glEnable(gl.GL_LIGHT2)

    gl.glMaterialfv(gl.GL_FRONT, gl.GL_AMBIENT, (0.1, 0.1, 0.1, 1.0))
    gl.glMaterialfv(gl.GL_FRONT, gl.GL_SHININESS, (10,))


def handle_videoresize_event(event):
    WINDOWED_SIZE[0] = event.w
    WINDOWED_SIZE[1] = event.h
    on_resize((event.w, event.h))


def on_resize(display_size):
    print("on_resize", display_size)
    global display
    global real_display_height, display_yoffset
    global browse_curve, header_curve  # ,screenshot_curve

    DisplayLists.clear()

    Render.display_width, Render.display_height = display_size  # display# .get_size()
    State.display_aspect = Render.display_width / Render.display_height
    print(WINDOWED_SIZE)
    Settings.windowed_size = tuple(WINDOWED_SIZE)

    real_display_height = Render.display_height
    # if Render.display_width / Render.display_height < 4 / 3:
    #    Render.display_height = int(Render.display_width / 4 * 3)
    #    display_yoffset = (real_display_height - Render.display_height) // 2
    # else:
    display_yoffset = 0
    # glViewport(0, 0, Render.display_width, real_display_height)
    # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    # swap_buffers()
    # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    # swap_buffers()
    print(display_yoffset)

    reinit_fonts()

    print(0, display_yoffset, Render.display_width, Render.display_height)
    gl.glViewport(
        0, display_yoffset, Render.display_width, Render.display_height)

    # aspect_ratio = max(4 / 3, (Render.display_width / Render.display_height))
    factor = (Render.display_width / Render.display_height) / (1024 / 600)
    browse_curve = Bezier.bezier(
        (-5.0 * factor, -10.0),
        (-1.7 * factor, -0.0),
        (1.7 * factor, -0.0),
        (5.0 * factor, -10.0)
    )
    header_curve = Bezier.bezier(
        (-2.0 * factor, 0.00),
        (-1.0 * factor, 0.075),
        (1.0 * factor, 0.075),
        (2.0 * factor, 0.00),
        steps=50
    )

    # if USE_MENU_TRANSITIONS:
    #    State.fbo = FrameBufferObject(
    #        Render.display_width, Render.display_height)
    Render.dirty = True 


class LogoStrength(object):
    anim = None
    value = 1.0


def set_ingame_status():
    # State.fade_start = State.time
    # State.fade_end = State.time + 86400.0 * 365.0
    # State.fade_from = (0.0, 0.0, 0.0, 0.0)
    # State.fade_to = (0.0, 0.0, 0.0, 0.0)
    # State.fade_splash = False
    print("State.currently_ingame = True")
    State.currently_ingame = True


def back_to_menu_from_game():
    print("State.currently_ingame = False")
    State.currently_ingame = False

    set_items_brightness(0.66, duration=0.500)

    RunTransition.value = 0.0
    RunTransition.anim = None
    LogoStrength.anim = None
    LogoStrength.value = 1.0

    Render.zoom = 1.0
    Render.offset_x = 0.0
    Render.offset_y = 0.0

    State.fade_splash = True
    State.fade_start = State.time
    State.fade_end = State.time + 2.0
    # Use 2.0 here to force full black for 1 second
    State.fade_from = (0.0, 0.0, 0.0, 2.0)
    State.fade_to = (0.0, 0.0, 0.0, 0.0)


def fade_quit():
    print("fade_quit")
    duration = 0.500
    alpha = 0.0
    start = get_current_time()
    while True:
        alpha = min(1.0, (get_current_time() - start) / duration)

        def render_func():
            render_screen()
            Render.ortho_perspective()
            fs_emu_blending(True)
            fs_emu_texturing(False)
            gl.glDisable(gl.GL_DEPTH_TEST)
            gl.glBegin(gl.GL_QUADS)
            gl.glColor4f(0.0, 0.0, 0.0, alpha)
            gl.glVertex2f(-10.0, -1.0)
            gl.glVertex2f(10.0, -1.0)
            gl.glVertex2f(10.0, 1.0)
            gl.glVertex2f(-10.0, 1.0)
            gl.glEnd()
            gl.glEnable(gl.GL_DEPTH_TEST)
            # fs_emu_blending(False)
            fs_emu_texturing(True)
            swap_buffers()
            Render.dirty = True

        Render.dirty = True
        main_loop_iteration(input_func=None, render_func=render_func)
        if alpha >= 1.0:
            break


def default_render_func():
    # rendered = False
    
    time_str = time.strftime("%H:%M")
    if time_str != Render.last_time_str:
        # clock must be updated, at least
        Render.dirty = True
        Render.last_time_str = time_str
    
    if Render.dirty or ALWAYS_RENDER:
        # print(Render.frame_number)
        render_screen()
        # rendered = True
        Render.twice = False
        # Render.twice = True
    else:
        if not Render.twice:
            Render.twice = True
            render_screen()
            # rendered = True
    if RENDER_DEBUG_SQUARES:
        render_debug_square_2()
    # if not rendered:
    #     pass
    # time.sleep(0.01)
    swap_buffers()


def main_loop_iteration(
        input_func=default_input_func, render_func=default_render_func):
    # if State.currently_ingame:
    #     print("currently ingame")
    #     #return False

    # print("main loop iteration")
    # time.sleep(0.1)
    # stop_loop = False

    # if State.idle_from and State.idle_from < get_current_time():
    #     #print(State.idle_from)
    #     if not Render.dirty:
    #         #print("waiting for events...")
    #         events = [pygame.event.wait()]
    #     else:
    #         events = pygame.event.get()
    # else:
    #     events = pygame.event.get()

    State.time = get_current_time()

    Main.process()

    # t = State.time

    # if len(char_buffer) > 2 and t - char_buffer_last_updated > 0.5:
    # print(current_menu, isinstance(current_menu, ItemMenu))
    # if isinstance(current_menu, ItemMenu):
    #
    #    # if len(char_buffer) > 2 and t - char_buffer_last_updated > 0.5:
    #    if len(char_buffer) > 2:
    #        create_search_results_menu(char_buffer)
    #    if len(char_buffer) <= 2:
    #        # collapse search menu
    #
    #        if hasattr(current_menu, "search_text"):
    #            print("setting current_menu (something with search)")
    #            set_current_menu(current_menu.parent_menu)

    if State.hide_mouse_time and State.time > State.hide_mouse_time:
        if Mouse.focus:
            # keep cursor visible if mouse has focus
            pass
        else:
            State.hide_mouse_time = 0
            Mouse.set_visible(False)

    # idle_events_only = True

    for event in InputHandler.pop_all_events():

        # if event.type != IDLE_EVENT:
        #     idle_events_only = False
        #
        # if event.type == pygame.MOUSEMOTION:
        #     if State.time - State.start_time < 1.0:
        #         # ignore initial events
        #         pass
        #     elif not State.mouse_visible:
        #         Mouse.set_visible(True)
        #     State.hide_mouse_time = State.time + 0.250
        #     mouse_x = event.pos[0] / Render.display_width * 1920
        #     mouse_y = 1080 - event.pos[1] / Render.display_height * 1080
        #     last_focus = Mouse.focus
        #     for item in Mouse.items:
        #         if not item.enabled:
        #             continue
        #         if item.x <= mouse_x <= item.x + item.w and \
        #                 item.y <= mouse_y <= item.y + item.h:
        #             Mouse.focus = item
        #             break
        #     else:
        #         Mouse.focus = None
        #     if last_focus != Mouse.focus:
        #         Render.dirty = True
        #
        # elif event.type == pygame.MOUSEBUTTONDOWN:
        #     if State.mouse_visible:
        #         if Mouse.focus:
        #             Mouse.focus.activate(State.current_menu)
        #             Render.dirty = True
        #
        # if event.type == pygame.QUIT:
        #     #command = pyapp.user.ini.get("Command/Shutdown", "").strip()
        #     #if command:
        #     #    # Do not allow ESC to quit if this causes a shutdown
        #     #    pass
        #     #else:
        #     print("State.quit = True")
        #     State.quit = True

        InputHandler.handle_event(event)

        # if event.type == pygame.VIDEORESIZE:
        #     handle_videoresize_event(event)
        # elif event.type == pygame.VIDEOEXPOSE:
        #     Render.dirty = True
        #
        # if event["type"] == "key-down":
        #     # hack to make this work only when using the default
        #     # input func
        #     if input_func == default_input_func:
        #         if event.key == pygame.K_F10:
        #             if event.mod & pygame.KMOD_LCTRL:
        #                 open_terminal()
        #         if event.key == pygame.K_F5:
        #             rescan_games()
        #         if event.key == pygame.K_F6:
        #             Render.display_fps = not Render.display_fps
        #             Render.dirty = True
        #             Notification("Display FPS:\n" + (
        #                 "Enabled" if Render.display_fps else "Disabled"))
        #         #         #elif char_buffer and event.key == pygame.K_RETURN:
        #         #    if character_press("RETURN"):
        #         #        # was handled
        #         #        InputHandler.get_button() # clear OK status
        #         #elif current_game is None and char_buffer and event.key == pygame.K_SPACE:
        #         #    InputHandler.get_button() # clear OK status
        #         #    character_press(" ")
        #         #elif current_game is None and char_buffer and event.key == pygame.K_BACKSPACE:
        #         #    InputHandler.get_button() # clear OK status
        #         #    character_press("BACKSPACE")
        #         elif event.unicode and event.unicode in SEARCH_CHARS:
        #         #elif event.unicode:
        #             #if event.key < pygame.K_KP0 or event.key > pygame.K_KP9:
        #             #    #InputHandler.get_button() # clear SHIFT/button status
        #             #    character_press(event.unicode)

        if event["type"] == "text":
            # if key was handled as a virtual button, only allow this
            # character if already started typing something, or else
            # navigating with X-Arcade may start searching for games
            if InputHandler.peek_button():
                if len(current_menu.search_text) > 0:
                    character_press(event["text"])
                    # reset InputHandler so that virtual button
                    # is not processed, since we handled this press
                    # as a char
                    InputHandler.get_button()
            else:
                character_press(event["text"])

    AnimationSystem.update()
    NotificationRender.update()
    #
    # if idle_events_only:
    #     # do not update State.idle_from
    #     pass
    # elif len(events) > 0:
    #     #print(events)
    #     if IDLE:
    #         State.idle_from = State.time + IDLE
    #         #print(State.idle_from)

    if Render.dirty or ALWAYS_RENDER:
        if Render.non_dirty_state:
            print("must start rendering again...")
        Render.non_dirty_state = False
        render_func()
    else:
        # if not Render.non_dirty_state:
        #     print("pause rendering!")
        # Render.non_dirty_state = True
        Render.dirty = True

    button = InputHandler.get_button()
    if button:
        print("InputHandler.get_button", button, InputHandler.last_device)
        GameCenter.register_user_activity()
        if input_func:
            input_func(button)
        Render.dirty = True
    if InputHandler.repeat_info:
        Render.dirty = True

    # if IDLE:
    if AnimationSystem.is_active():
        # State.idle_from = None
        Render.dirty = True
        # elif State.idle_from is None:
        #     State.idle_from = State.time + IDLE
    if not IDLE:
        Render.dirty = True
    # except KeyboardInterrupt:
    #     print "KeyboardInterrupt"
    #     return
    # return stop_loop

    if not Render.display_sync:
        t = time.time()
        diff = t - Render.display_last_iteration
        # print(diff)
        frame_time = 1 / 60.0
        if diff < frame_time:
            # FIXME: use real refresh rate / frame time
            sleep_time = frame_time - diff
            time.sleep(sleep_time)
        Render.display_last_iteration = t

    return State.quit



'''OpenGL extension NV.vertex_attrib_integer_64bit

This module customises the behaviour of the 
OpenGL.raw.GL.NV.vertex_attrib_integer_64bit to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension provides support for specifying vertex attributes with
	64-bit integer components, analagous to the 64-bit floating point support
	added in EXT_vertex_attrib_64bit.
	
	Additionally, it provides the VertexAttribLFormatNV entry point to specify
	bindless vertex attribute arrays with 64-bit integer or floating-point
	components in conjunction with the NV_vertex_buffer_unified_memory
	extension.
	

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/NV/vertex_attrib_integer_64bit.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.NV.vertex_attrib_integer_64bit import *
from OpenGL.raw.GL.NV.vertex_attrib_integer_64bit import _EXTENSION_NAME

def glInitVertexAttribInteger64BitNV():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


def sort_configurations(items):
    # Cracked games score better, since these need no
    # manual lookup, etc.
    score_table = [
        [" AGA ", -(2 ** 16)],   # amiga AGA version
        [" cr ", -(2 ** 15)],    # cracked version
        [" NTSC ", -(2 ** 14)],  # prefer NTSC version
    ]
    scored = []
    for item in items:
        check = ' ' + item.replace(',', ' ') + ' '
        score = 0
        for p, s in score_table:
            if p in check:
                # scored.append((0, item))
                score += s
        # prefer shortest configs, all else being equal
        score += len(item)
        scored.append((score, item))
        # scored.append(1, item)
    scored.sort()
    return [x[1] for x in scored]



from fsgs.platform import PlatformHandler
from fsgs.mednafen.game_boy_color import GameBoyColorRunner
from .loader import SimpleLoader


class GameBoyColorPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Game Boy Color"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return GameBoyColorLoader(fsgs)

    def get_runner(self, fsgs):
        return GameBoyColorRunner(fsgs)


class GameBoyColorLoader(SimpleLoader):
    pass



import fsbc.Application
import sys

# if "arcade" in sys.argv[0] or len(sys.argv) > 1 and "arcade" in sys.argv[1]:
#     import_order = ["PyQT4", "PySide", "PyQT5"]
# else:
#     import_order = ["PyQT5", "PyQT4", "PySide"]
# import_order = ["PyQT4", "PySide", "PyQT5"]

import_order = ["PyQT5"]

if "--qt-bindings=pyqt5" in sys.argv:
    import_order = ["PyQT5"]
elif "--qt-bindings=pyqt4" in sys.argv:
    import_order = ["PyQT4"]
elif "--qt-bindings=pyside" in sys.argv:
    import_order = ["PySide"]

for name in import_order:
    if name == "PyQT5":
        try:
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            import PyQt5
        except ImportError:
            continue
        else:
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt5.QtCore import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt5.QtGui import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt5.QtWidgets import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt5.QtCore import pyqtSignal as Signal
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt5.QtCore import pyqtSignal as QSignal
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt5.QtOpenGL import *
            try:
                # noinspection PyUnresolvedReferences, PyPackageRequirements
                from PyQt5.QtX11Extras import *
            except ImportError:
                pass
    elif name == "PyQT4":
        try:
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            import PyQt4
        except ImportError:
            continue
        else:
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            import sip
            sip.setapi("QString", 2)
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt4.QtCore import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt4.QtGui import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt4.QtCore import pyqtSignal as Signal
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt4.QtCore import pyqtSignal as QSignal
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PyQt4.QtOpenGL import *
    elif name == "PySide":
        try:
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            import PySide
        except ImportError:
            continue
        else:
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PySide.QtCore import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PySide.QtCore import Signal as QSignal
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PySide.QtGui import *
            # noinspection PyUnresolvedReferences, PyPackageRequirements
            from PySide.QtOpenGL import *
    else:
        raise Exception("unknown QT python bindings specified")
    break
else:
    raise Exception("no QT python bindings found")



from fsui.qt import QLineEdit, Signal
from .Widget import Widget


class TextField(QLineEdit, Widget):

    changed = Signal()
    activated = Signal()

    def __init__(self, parent, text="", read_only=False):
        QLineEdit.__init__(self, text, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        self.setReadOnly(read_only)
        self.textChanged.connect(self.__text_changed)
        self.returnPressed.connect(self.__return_pressed)

    def get_text(self):
        return self.text()

    def set_text(self, text):
        self.setText(text)

    def set_cursor_position(self, position):
        self.setCursorPosition(position)

    def on_change(self):
        pass

    def __text_changed(self, text):
        self.changed.emit()
        self.on_change()

    def select_all(self):
        self.selectAll()

    def __return_pressed(self):
        self.activated.emit()


class PasswordField(TextField):

    def __init__(self, parent, text=""):
        TextField.__init__(self, parent, text)
        self.setEchoMode(QLineEdit.Password)



import json
import sqlite3
from binascii import unhexlify
import zlib
from .BaseDatabase import BaseDatabase


VERSION = 18
RESET_VERSION = 18
DUMMY_UUID = b"'\\x8b\\xbb\\x00Y\\x8bqM\\x15\\x972\\xa8-t_\\xb2\\xfd'"


class GameDatabase(BaseDatabase):

    def __init__(self, path):
        BaseDatabase.__init__(self, BaseDatabase.SENTINEL)
        self._path = path
        # self._connection = None
        # self._cursor = None

    def get_path(self):
        return self._path

    def get_version(self):
        return VERSION

    def get_reset_version(self):
        return RESET_VERSION

    @staticmethod
    def query(q):
        return q.replace("%s", "?")

    def get_last_game_id(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT max(id) FROM game")
        row = cursor.fetchone()
        return row[0] or 0

    def get_ratings_for_game(self, game_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT like_rating, work_rating FROM rating WHERE "
            "game_uuid = ?", (game_uuid,))
        row = cursor.fetchone()
        if row is not None:
            return row[0], row[1]
        else:
            return 0, 0

    def get_license_code_for_url(self, url):
        # cursor = self.internal_cursor()
        # cursor.execute(
        #     "SELECT license_code FROM file "
        #     "WHERE external = ? LIMIT 1", (url,))
        # row = cursor.fetchone()
        # if not row:
        #     return None
        # return row[0]
        return None

    def add_game(self, game_id, game_uuid, game_data):
        # print("add game", repr(game_id), repr(game_uuid), repr(game_data))
        cursor = self.internal_cursor()
        cursor.execute(
            "DELETE FROM game WHERE uuid = ?",
            (sqlite3.Binary(game_uuid),))
        cursor.execute(
            "INSERT INTO game (id, uuid, data) VALUES (?, ?, ?)",
            (game_id, sqlite3.Binary(game_uuid), sqlite3.Binary(game_data)))

    def delete_game(self, game_id, game_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "DELETE FROM game WHERE uuid = ?",
            (sqlite3.Binary(game_uuid),))
        cursor.execute(
            "DELETE FROM game WHERE uuid = ?",
            (sqlite3.Binary(DUMMY_UUID),))
        cursor.execute(
            "INSERT INTO game (id, uuid, data) VALUES (?, ?, ?)",
            (game_id, sqlite3.Binary(DUMMY_UUID), ""))

    def get_game_values_for_uuid(self, game_uuid, recursive=True):
        print("get_game_values_for_uuid", game_uuid)
        assert game_uuid
        assert isinstance(game_uuid, str)
        # cursor = self.internal_cursor()
        # query = "SELECT id FROM game WHERE uuid = ?"
        # args = (sqlite3.Binary(unhexlify(game_uuid.replace("-", ""))),)
        # cursor.execute(query, args)
        # game_id = cursor.fetchone()[0]
        # return self.get_game_values(game_id, recursive)
        return self.get_game_values(game_uuid=game_uuid, recursive=recursive)

    def get_game_values(self, game_id=None, *, game_uuid=None, recursive=True):
        cursor = self.internal_cursor()
        if game_uuid is not None:
            cursor.execute(
                "SELECT data FROM game WHERE uuid = ?",
                (sqlite3.Binary(unhexlify(game_uuid.replace("-", ""))),))
        else:
            cursor.execute(
                "SELECT data FROM game WHERE id = ?", (game_id,))
        data = zlib.decompress(cursor.fetchone()[0])
        data = data.decode("UTF-8")
        doc = json.loads(data)
        next_parent_uuid = doc.get("parent_uuid", "")
        while next_parent_uuid and recursive:
            # treat game_uuid special, it will be the first parent_uuid
            # in the chain
            doc["game_uuid"] = next_parent_uuid
            cursor.execute(
                "SELECT data FROM game WHERE uuid = ?",
                (sqlite3.Binary(
                    unhexlify(next_parent_uuid.replace("-", ""))),))
            row = cursor.fetchone()
            if not row:
                raise Exception("could not find parent {0}".format(
                    next_parent_uuid))
            data = zlib.decompress(row[0])
            data = data.decode("UTF-8")
            next_doc = json.loads(data)
            next_parent_uuid = next_doc.get("parent_uuid", "")
            # let child doc overwrite and append values to parent doc
            next_doc.update(doc)
            doc = next_doc
        return doc

    def get_game_database_version(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT database_version FROM metadata")
        return cursor.fetchone()[0]

    def set_game_database_version(self, version):
        cursor = self.internal_cursor()
        cursor.execute("UPDATE metadata SET database_version = ?", (version,))

    def clear(self):
        cursor = self.internal_cursor()
        cursor.execute("DELETE FROM rating")
        cursor.execute("DELETE FROM game")

    def update_database_to_version_18(self):
        cursor = self.internal_cursor()
        cursor.execute("""
            ALTER TABLE metadata ADD COLUMN games_version
            INTEGER NOT NULL DEFAULT 0;
        """)
        cursor.execute("""
            CREATE TABLE game (
                id INTEGER PRIMARY KEY,
                uuid BLOB,
                data BLOB
            );
        """)
        cursor.execute("""
            CREATE INDEX game_uuid ON game (uuid);
        """)
        cursor.execute("""
            CREATE TABLE rating (
                game_uuid VARCHAR(36) PRIMARY KEY NOT NULL,
                work_rating INT NOT NULL,
                like_rating INT NOT NULL,
                updated TIMESTAMP NULL
            );
        """)
        cursor.execute("""
            ALTER TABLE metadata ADD COLUMN database_version
            INTEGER NOT NULL DEFAULT 0;
        """)



import sys

windows = sys.platform == "win32"
linux = sys.platform.startswith("linux")
macosx = sys.platform == "darwin"

if windows:
    platform = "windows"
elif linux:
    platform = "linux"
elif macosx:
    platform = "macosx"
else:
    platform = "unknown"



# Automatically generated - do not edit by hand


class Option(object):
    """Constants for option names."""    
    ACCELERATOR = "accelerator"
    ACCELERATOR_MEMORY = "accelerator_memory"
    AUDIO_BUFFER_TARGET_BYTES = "audio_buffer_target_bytes"
    AUDIO_BUFFER_TARGET_SIZE = "audio_buffer_target_size"
    AUDIO_FREQUENCY = "audio_frequency"
    AUTOMATIC_INPUT_GRAB = "automatic_input_grab"
    BLIZZARD_SCSI_KIT = "blizzard_scsi_kit"
    BSDSOCKET_LIBRARY = "bsdsocket_library"
    BUILTIN_CONFIGS = "builtin_configs"
    CDROM_DRIVE_COUNT = "cdrom_drive_count"
    CHIP_MEMORY = "chip_memory"
    CONFIG_FEATURE = "config_feature"
    CPU = "cpu"
    CPU_IDLE = "cpu_idle"
    DATABASE_ARCADE = "database_arcade"
    DATABASE_AUTH = "database_auth"
    DATABASE_DOS = "database_dos"
    DATABASE_EMAIL = "database_email"
    DATABASE_FEATURE = "database_feature"
    DATABASE_GBA = "database_gba"
    DATABASE_LOCKER = "database_locker"
    DATABASE_NES = "database_nes"
    DATABASE_PASSWORD = "database_password"
    DATABASE_SERVER = "database_server"
    DATABASE_SHOW_ADULT = "database_show_adult"
    DATABASE_SHOW_GAMES = "database_show_games"
    DATABASE_SNES = "database_snes"
    DATABASE_USERNAME = "database_username"
    DEVICE_ID = "device_id"
    DONGLE_TYPE = "dongle_type"
    FADE_IN_DURATION = "fade_in_duration"
    FLOPPY_DRIVE_COUNT = "floppy_drive_count"
    FLOPPY_DRIVE_SPEED = "floppy_drive_speed"
    FLOPPY_DRIVE_VOLUME = "floppy_drive_volume"
    FLOPPY_DRIVE_VOLUME_EMPTY = "floppy_drive_volume_empty"
    FSAA = "fsaa"
    FULLSCREEN = "fullscreen"
    GRAPHICS_CARD = "graphics_card"
    GRAPHICS_MEMORY = "graphics_memory"
    INITIAL_INPUT_GRAB = "initial_input_grab"
    IRC_NICK = "irc_nick"
    IRC_SERVER = "irc_server"
    KEEP_ASPECT = "keep_aspect"
    KEYBOARD_INPUT_GRAB = "keyboard_input_grab"
    KEYBOARD_KEY_BACKSLASH = "keyboard_key_backslash"
    KEYBOARD_KEY_EQUALS = "keyboard_key_equals"
    KEYBOARD_KEY_INSERT = "keyboard_key_insert"
    KEYBOARD_KEY_LESS = "keyboard_key_less"
    KICKSTART_SETUP = "kickstart_setup"
    LAUNCHER_FONT_SIZE = "launcher_font_size"
    LAUNCHER_THEME = "launcher_theme"
    LOAD_STATE = "load_state"
    LOG_AUTOSCALE = "log_autoscale"
    LOG_BSDSOCKET = "log_bsdsocket"
    LOG_INPUT = "log_input"
    LOG_QUERY_PLANS = "log_query_plans"
    LOW_LATENCY_VSYNC = "low_latency_vsync"
    MIDDLE_CLICK_UNGRAB = "middle_click_ungrab"
    MIN_FIRST_LINE_NTSC = "min_first_line_ntsc"
    MIN_FIRST_LINE_PAL = "min_first_line_pal"
    MONITOR = "monitor"
    MOUSE_SPEED = "mouse_speed"
    NETPLAY_FEATURE = "netplay_feature"
    NETPLAY_TAG = "netplay_tag"
    RTG_SCANLINES = "rtg_scanlines"
    SCANLINES = "scanlines"
    SOUND_CARD = "sound_card"
    STEREO_SEPARATION = "stereo_separation"
    SWAP_CTRL_KEYS = "swap_ctrl_keys"
    TEXTURE_FILTER = "texture_filter"
    TEXTURE_FORMAT = "texture_format"
    UAE_A2065 = "uae_a2065"
    UAE_A3000MEM_SIZE = "uae_a3000mem_size"
    UAE_CHIPSET_COMPATIBLE = "uae_chipset_compatible"
    UAE_CPU_FREQUENCY = "uae_cpu_frequency"
    UAE_CPU_MULTIPLIER = "uae_cpu_multiplier"
    UAE_CPU_SPEED = "uae_cpu_speed"
    UAE_CPU_THROTTLE = "uae_cpu_throttle"
    UAE_FASTMEM2_SIZE = "uae_fastmem2_size"
    UAE_FASTMEM_AUTOCONFIG = "uae_fastmem_autoconfig"
    UAE_FASTMEM_SIZE = "uae_fastmem_size"
    UAE_FORCE_0X10000000_Z3 = "uae_force_0x10000000_z3"
    UAE_GFX_LINEMODE = "uae_gfx_linemode"
    UAE_MBRESMEM_SIZE = "uae_mbresmem_size"
    UAE_RTC = "uae_rtc"
    UAE_SANA2 = "uae_sana2"
    UAE_SLIRP_IMPLEMENTATION = "uae_slirp_implementation"
    UAE_SOUND_OUTPUT = "uae_sound_output"
    UAE_TOCCATA = "uae_toccata"
    UAE_Z3CHIPMEM_SIZE = "uae_z3chipmem_size"
    UAE_Z3MAPPING = "uae_z3mapping"
    UAE_Z3MEM2_SIZE = "uae_z3mem2_size"
    UAE_Z3MEM_SIZE = "uae_z3mem_size"
    UAEGFX_CARD = "uaegfx_card"
    UAEM_WRITE_FLAGS = "uaem_write_flags"
    VIDEO_FORMAT = "video_format"
    VIDEO_SYNC = "video_sync"
    VIDEO_SYNC_METHOD = "video_sync_method"
    VOLUME = "volume"
    WHDLOAD_SPLASH_DELAY = "whdload_splash_delay"
    WINDOW_BORDER = "window_border"
    ZOOM = "zoom"


class Options(object):

    @staticmethod
    def get(name):
        return options[name]


N_ = lambda x: x


options = {

    "accelerator": {
        "default": "0",
        "description": N_("CPU Accelerator Board"),
        "type": "Choice",
        "values": [
            ("0", N_("None")),
            ("blizzard-1230-iv", "Blizzard 1230 IV"),
            ("blizzard-1240", "Blizzard 1240"),
            ("blizzard-1260", "Blizzard 1260"),
            ("blizzard-ppc", "Blizzard PPC"),
            ("cyberstorm-ppc", "CyberStorm PPC"),
        ]
    },

    "accelerator_memory": {
        "default": "",
        "description": N_("CPU Accelerator Board RAM"),
        "type": "Choice",
        "values": [
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("16384", "16 MB"),
            ("32768", "32 MB"),
            ("65536", "64 MB"),
            ("131072", "128 MB"),
            ("262144", "256 MB"),
        ]
    },

    "audio_buffer_target_bytes": {
        "default": "8192",
        "description": N_("Audio buffer target size (bytes)"),
        "type": "integer",
        "min": 1024,
        "max": 32768,
    },

    "audio_buffer_target_size": {
        "default": "40",
        "description": N_("Audio buffer target size (ms)"),
        "type": "integer",
        "min": 1,
        "max": 100,
    },

    "audio_frequency": {
        "default": "48000",
        "description": N_("Audio output frequency"),
        "type": "choice",
        "values": [
            ("48000", "48000 Hz"),
            ("44100", "44100 Hz"),
            ("22050", "22050 Hz"),
            ("11025", "11025 Hz"),
        ]
    },

    "automatic_input_grab": {
        "default": "1",
        "description": N_("Grab Input on Click"),
        "type": "boolean",
    },

    "blizzard_scsi_kit": {
        "default": "0",
        "description": N_("Blizzard SCSI Kit"),
        "type": "boolean",
    },

    "bsdsocket_library": {
        "default": "0",
        "description": N_("UAE bsdsocket.library"),
        "type": "boolean",
    },

    "builtin_configs": {
        "default": "1",
        "description": N_("Include built-in configurations"),
        "type": "boolean",
    },

    "cdrom_drive_count": {
        "default": "",
        "description": N_("CD-ROM Drive Count"),
        "type": "Choice",
        "values": [
            ("0", "0"),
            ("1", "1"),
        ]
    },

    "chip_memory": {
        "default": "",
        "description": N_("Chip RAM"),
        "type": "Choice",
    },

    "config_feature": {
        "default": "0",
        "description": N_(
            "Enable experimental config visualization (requires restart)"),
        "type": "boolean",
    },

    "cpu": {
        "default": "",
        "description": N_("CPU"),
        "type": "Choice",
        "values": [
            ("68000", "68000"),
            ("68010", "68010"),
            ("68EC020", "68EC020"),
            ("68020", "68020"),
            ("68EC030", "68EC030"),
            ("68030", "68030"),
            ("68EC040", "68EC040"),
            ("68LC040", "68LC040"),
            ("68040-NOMMU", "68040-NOMMU"),
            ("68040", "68040"),
            ("68EC060", "68EC060"),
            ("68LC060", "68LC060"),
            ("68060-NOMMU", "68060-NOMMU"),
            ("68060", "68060"),
        ]
    },

    "cpu_idle": {
        "default": "",
        "description": N_(
            "Relax host CPU usage when using fastest-possible CPU"),
        "type": "integer",
        "min": 0,
        "max": 10,
    },

    "database_arcade": {
        "default": "",
        "description": N_("Enable/disable use of the Arcade game database"),
        "type": "boolean",
    },

    "database_auth": {
        "default": "",
        "description": N_("Game database authentication"),
        "type": "string",
    },

    "database_dos": {
        "default": "",
        "description": N_("Enable/disable use of the DOS game database"),
        "type": "boolean",
    },

    "database_email": {
        "default": "",
        "description": N_("Game database email"),
        "type": "string",
    },

    "database_feature": {
        "default": "0",
        "description": N_("Enable online database support (requires restart)"),
        "type": "boolean",
    },

    "database_gba": {
        "default": "",
        "description": N_(
            "Enable/disable use of the Game Boy Advance database"),
        "type": "boolean",
    },

    "database_locker": {
        "default": "",
        "description": N_("Enable/disable use of OAGD.net locker"),
        "type": "boolean",
    },

    "database_nes": {
        "default": "",
        "description": N_("Enable/disable use of the Nintendo (NES) database"),
        "type": "boolean",
    },

    "database_password": {
        "default": "",
        "description": N_("Game database password"),
        "type": "string",
    },

    "database_server": {
        "default": "oagd.net",
        "description": N_("Game Database Server"),
        "type": "string",
    },

    "database_show_adult": {
        "default": "0",
        "description": N_("Adult-Themed Games"),
        "type": "boolean",
    },

    "database_show_games": {
        "default": "1",
        "description": N_("Database Games"),
        "type": "choice",
        "values": [
            ("0", N_("All Games")),
            ("1", N_("Available and Downloadable Games")),
            ("2", N_("Available and Auto-Downloadable Games")),
            ("3", N_("Available Games")),
        ]
    },

    "database_snes": {
        "default": "",
        "description": N_("Enable/disable use of the Super Nintendo database"),
        "type": "boolean",
    },

    "database_username": {
        "default": "",
        "description": N_("Game database user name"),
        "type": "string",
    },

    "device_id": {
        "default": "",
        "description": N_("Device ID used with OAGD.net authentication"),
        "type": "string",
    },

    "dongle_type": {
        "default": "",
        "description": N_("Dongle"),
        "type": "Choice",
        "values": [
            ("robocop 3", "robocop 3"),
            ("leaderboard", "leaderboard"),
            ("b.a.t. ii", "b.a.t. ii"),
            ("italy'90 soccer", "italy'90 soccer"),
            ("dames grand maitre", "dames grand maitre"),
            ("rugby coach", "rugby coach"),
            ("cricket captain", "cricket captain"),
            ("leviathan", "leviathan"),
        ]
    },

    "fade_in_duration": {
        "default": "0",
        "description": N_("Fade-in Duration on Start"),
        "type": "",
    },

    "floppy_drive_count": {
        "default": "",
        "description": N_("Floppy Drive Count"),
        "type": "Choice",
        "values": [
            ("0", "0"),
            ("1", "1"),
            ("2", "2"),
            ("3", "3"),
            ("4", "4"),
        ]
    },

    "floppy_drive_speed": {
        "default": "100",
        "description": N_("Floppy Drive Speed"),
        "type": "Choice",
        "values": [
            ("0", N_("Turbo")),
            ("100", "100%"),
            ("200", "200 %"),
            ("400", "400 %"),
            ("800", "800 %"),
        ]
    },

    "floppy_drive_volume": {
        "default": "67",
        "description": N_("Floppy Drive Volume"),
        "type": "integer",
        "min": 0,
        "max": 100,
    },

    "floppy_drive_volume_empty": {
        "default": "67",
        "description": N_("Empty Floppy Drive Volume"),
        "type": "integer",
        "min": 0,
        "max": 100,
    },

    "fsaa": {
        "default": "0",
        "description": N_("Full-scene anti-aliasing (FSAA)"),
        "type": "choice",
        "values": [
            ("0", N_("Off")),
            ("2", "2x2"),
            ("4", "4x4"),
            ("8", "8x8"),
        ]
    },

    "fullscreen": {
        "default": "0",
        "description": N_("Start FS-UAE in fullscreen mode"),
        "type": "boolean",
    },

    "graphics_card": {
        "default": "none",
        "description": N_("Graphics Card"),
        "type": "Choice",
        "values": [
            ("none", N_("None")),
            ("uaegfx", "UAEGFX"),
            ("uaegfx-z2", "UAEGFX Zorro II"),
            ("uaegfx-z3", "UAEGFX Zorro III"),
            ("picasso-ii", "Picasso II Zorro II"),
            ("picasso-ii+", "Picasso II+ Zorro II"),
            ("picasso-iv", "Picasso IV"),
            ("picasso-iv-z2", "Picasso IV Zorro II"),
            ("picasso-iv-z3", "Picasso IV Zorro III"),
        ]
    },

    "graphics_memory": {
        "default": "",
        "description": N_("Graphics Card VRAM"),
        "type": "Choice",
        "values": [
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("16384", "16 MB"),
            ("32768", "32 MB"),
            ("65536", "64 MB"),
            ("131072", "128 MB"),
            ("262144", "256 MB"),
        ]
    },

    "initial_input_grab": {
        "default": "",
        "description": N_("Grab Input on FS-UAE Startup"),
        "type": "boolean",
    },

    "irc_nick": {
        "default": "",
        "description": N_("IRC nickname"),
        "type": "string",
    },

    "irc_server": {
        "default": "irc.fengestad.no",
        "description": N_("Custom IRC server address"),
        "type": "string",
    },

    "keep_aspect": {
        "default": "0",
        "description": N_("Keep aspect ratio when scaling (do not stretch)"),
        "type": "boolean",
    },

    "keyboard_input_grab": {
        "default": "1",
        "description": N_("Grab keyboard when input is grabbed"),
        "type": "boolean",
    },

    "keyboard_key_backslash": {
        "default": "action_key_backslash",
        "description": N_("Host Key BACKSLASH"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ]
    },

    "keyboard_key_equals": {
        "default": "action_key_equals",
        "description": N_("Host Key EQUALS"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ]
    },

    "keyboard_key_insert": {
        "default": "action_key_2b",
        "description": N_("Host Key INSERT"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ]
    },

    "keyboard_key_less": {
        "default": "action_key_30",
        "description": N_("Host Key LESS"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ]
    },

    "kickstart_setup": {
        "default": "1",
        "description": N_(
            "Show kickstart setup page on startup when all ROMs are missing"),
        "type": "boolean",
    },

    "launcher_font_size": {
        "default": "",
        "description": N_("Font Size"),
        "type": "Integer",
        "min": 6,
        "max": 18,
    },

    "launcher_theme": {
        "default": "fusion",
        "description": N_("Theme"),
        "type": "Choice",
        "values": [
            ("standard", N_("Standard")),
            ("fusion", "Fusion Auto"),
            ("fusion-plain", "Fusion Plain"),
            ("fusion-adwaita", "Fusion Adwaita"),
            ("fusion-dark", "Fusion Dark"),
        ]
    },

    "load_state": {
        "default": "",
        "description": N_("Load state by number"),
        "type": "integer",
        "min": 1,
        "max": 9,
    },

    "log_autoscale": {
        "default": "0",
        "description": N_("Log Autoscale Changes"),
        "type": "boolean",
    },

    "log_bsdsocket": {
        "default": "0",
        "description": "",
        "type": "boolean",
    },

    "log_input": {
        "default": "0",
        "description": N_("Log Input Events"),
        "type": "boolean",
    },

    "log_query_plans": {
        "default": "",
        "description": N_("Log database query plans"),
        "type": "",
    },

    "low_latency_vsync": {
        "default": "0",
        "description": N_("Low latency video sync"),
        "type": "boolean",
    },

    "middle_click_ungrab": {
        "default": "1",
        "description": N_("Ungrab Input on Middle Mouse Button"),
        "type": "boolean",
    },

    "min_first_line_ntsc": {
        "default": "21",
        "description": N_("First rendered line (NTSC)"),
        "type": "",
    },

    "min_first_line_pal": {
        "default": "26",
        "description": N_("First rendered line (PAL)"),
        "type": "",
    },

    "monitor": {
        "default": "middle-left",
        "description": N_("Monitor to display FS-UAE on (fullscreen)"),
        "type": "choice",
        "values": [
            ("left", N_("Left")),
            ("middle-left", N_("Middle Left")),
            ("middle-right", N_("Middle Right")),
            ("right", N_("Right")),
        ]
    },

    "mouse_speed": {
        "default": "100",
        "description": N_("Mouse Speed (%)"),
        "type": "integer",
        "min": 1,
        "max": 500,
    },

    "netplay_feature": {
        "default": "0",
        "description": N_(
            "Enable experimental net play GUI (requires restart)"),
        "type": "boolean",
    },

    "netplay_tag": {
        "default": "UNK",
        "description": N_("Net play tag (max 3 characters)"),
        "type": "string",
    },

    "rtg_scanlines": {
        "default": "0",
        "description": N_("Render scan lines in RTG graphics mode"),
        "type": "boolean",
    },

    "scanlines": {
        "default": "0",
        "description": N_("Render scan lines"),
        "type": "boolean",
    },

    "sound_card": {
        "default": "0",
        "description": N_("Sound Card"),
        "type": "Choice",
        "values": [
            ("0", N_("None")),
            ("toccata", "Toccata"),
        ]
    },

    "stereo_separation": {
        "default": "70",
        "description": N_("Stereo Separation"),
        "type": "choice",
        "values": [
            ("100", "100%"),
            ("90", "90%"),
            ("80", "80%"),
            ("70", "70%"),
            ("60", "60%"),
            ("50", "50%"),
            ("40", "40%"),
            ("30", "30%"),
            ("20", "20%"),
            ("10", "10%"),
            ("0", "0%"),
        ]
    },

    "swap_ctrl_keys": {
        "default": "0",
        "description": N_("Swap left and right CTRL keys"),
        "type": "boolean",
    },

    "texture_filter": {
        "default": "linear",
        "description": N_("Texture filter"),
        "type": "choice",
        "values": [
            ("nearest", "GL_NEAREST"),
            ("linear", "GL_LINEAR"),
        ]
    },

    "texture_format": {
        "default": "",
        "description": N_("Video texture format (on the GPU)"),
        "type": "choice",
        "values": [
            ("rgb", "GL_RGB"),
            ("rgb8", "GL_RGB8"),
            ("rgba", "GL_RGBA"),
            ("rgba8", "GL_RGBA8"),
            ("rgb5", "GL_RGB5"),
            ("rgb5_1", "GL_RGB5_1"),
        ]
    },

    "uae_a2065": {
        "default": "",
        "description": "uae_a2065",
        "type": "",
        "values": [
            ("slirp", "slirp"),
        ]
    },

    "uae_a3000mem_size": {
        "default": "",
        "description": "Size in megabytes of motherboard fast memory",
        "type": "integer",
        "min": 0,
        "max": 65536,
    },

    "uae_chipset_compatible": {
        "default": "",
        "description":
            "Enable default chipset features for a specific model",
        "type": "choice",
        "values": [
            ("-", "-"),
            ("Generic", "Generic"),
            ("CDTV", "CDTV"),
            ("CD32", "CD32"),
            ("A500", "A500"),
            ("A500+", "A500+"),
            ("A600", "A600"),
            ("A1000", "A1000"),
            ("A1200", "A1200"),
            ("A2000", "A2000"),
            ("A3000", "A3000"),
            ("A3000T", "A3000T"),
            ("A4000", "A4000"),
            ("A4000T", "A4000T"),
        ]
    },

    "uae_cpu_frequency": {
        "default": "",
        "description":
            "Specify the frequency of the emulated CPU in cycle-exact modes",
        "type": "float",
        "min": 1.0,
        "max": 100.0,
    },

    "uae_cpu_multiplier": {
        "default": "",
        "description": "FIXME",
        "type": "integer",
        "min": 0,
        "max": 256,
    },

    "uae_cpu_speed": {
        "default": "",
        "description": "Enable/disable fastest possible CPU speed",
        "type": "choice",
        "values": [
            ("real", "Approximate A500/A1200 or cycle-exact"),
            ("max", "Fastest possible"),
        ]
    },

    "uae_cpu_throttle": {
        "default": "",
        "description": "FIXME",
        "type": "float",
        "min": -900.0,
        "max": 5000.0,
    },

    "uae_fastmem2_size": {
        "default": "0",
        "description":
            "Size in MB of Zorro-II Fast RAM (second) expansion board",
        "type": "choice",
        "values": [
            ("0", "0"),
            ("1", "1"),
            ("2", "2"),
            ("4", "4"),
        ]
    },

    "uae_fastmem_autoconfig": {
        "default": "1",
        "description": "Autoconfig Z2 Fast RAM",
        "type": "boolean",
    },

    "uae_fastmem_size": {
        "default": "0",
        "description": "Size in MB of Zorro-II Fast RAM expansion board",
        "type": "choice",
        "values": [
            ("0", "0"),
            ("1", "1"),
            ("2", "2"),
            ("4", "4"),
            ("8", "8"),
        ]
    },

    "uae_force_0x10000000_z3": {
        "default": "false",
        "description": "Force Zorro-III address space at 0x10000000",
        "type": "boolean",
    },

    "uae_gfx_linemode": {
        "default": "",
        "description":
            "Controls how lines are doubled and interlaced modes are handled",
        "type": "choice",
        "values": [
            ("none", "Single / Single"),
            ("double", "Double / Double Frames"),
            ("double2", "Double / Double Fields"),
            ("double3", "Double / Double Fields+"),
            ("scanlines", "Scanlines / Double Frames"),
            ("scanlines2", "Scanlines / Double Fields"),
            ("scanlines3", "Scanlines / Double Fields+"),
            ("scanlines2p", "Double Fields / Double Frames"),
            ("scanlines2p2", "Double Fields / Double Fields"),
            ("scanlines2p3", "Double Fields / Double Fields+"),
            ("scanlines3p", "Double Fields+ / Double Frames"),
            ("scanlines3p2", "Double Fields+ / Double Fields"),
            ("scanlines3p3", "Double Fields+ / Double Fields+"),
        ]
    },

    "uae_mbresmem_size": {
        "default": "",
        "description": "Size in megabytes of processor slot fast memory",
        "type": "integer",
        "min": 0,
        "max": 131072,
    },

    "uae_rtc": {
        "default": "auto",
        "description": "Enable a real time clock (RTC) module.",
        "type": "Choice",
        "values": [
            ("auto", "Auto"),
            ("none", "None"),
            ("MSM6242B", "Oki MSM6242B"),
            ("RP5C01A", "Ricoh RP5C01A"),
            ("MSM6242B_A2000", "A2000 MSM6242B"),
        ]
    },

    "uae_sana2": {
        "default": "false",
        "description": "uae_sana2",
        "type": "BooleanUAE,",
    },

    "uae_slirp_implementation": {
        "default": "auto",
        "description": "Slirp Implementation",
        "type": "Choice",
        "values": [
            ("auto", "auto"),
            ("none", "none"),
            ("builtin", "builtin"),
            ("qemu", "qemu"),
        ]
    },

    "uae_sound_output": {
        "default": "",
        "description": "Sound emulation",
        "type": "",
        "values": [
            ("none", "Disabled"),
            ("interrupts", "Emulated, No Output"),
            ("exact", "Enabled"),
        ]
    },

    "uae_toccata": {
        "default": "",
        "description": "Toccata Z2 sound card emulation",
        "type": "uaeyesno",
    },

    "uae_z3chipmem_size": {
        "default": "",
        "description": "",
        "type": "integer",
    },

    "uae_z3mapping": {
        "default": "auto",
        "description": "JIT Direct compatible Z3 memory mapping",
        "type": "choice",
    },

    "uae_z3mem2_size": {
        "default": "",
        "description": "",
        "type": "integer",
    },

    "uae_z3mem_size": {
        "default": "",
        "description": "Size in MB of Zorro-III Fast RAM expansion board",
        "type": "integer",
    },

    "uaegfx_card": {
        "default": "0",
        "description": N_("uaegfx_card"),
        "type": "boolean",
    },

    "uaem_write_flags": {
        "default": "1",
        "description": N_("Write .uaem metadata files"),
        "type": "flags",
    },

    "video_format": {
        "default": "bgra",
        "description": N_("Video buffer format and color depth"),
        "type": "choice",
        "values": [
            ("bgra", N_("32-bit BGRA")),
            ("rgba", N_("32-bit RGBA")),
            ("rgb565", N_("16-bit")),
        ]
    },

    "video_sync": {
        "default": "off",
        "description": N_("Video synchronization"),
        "type": "choice",
        "values": [
            ("auto", N_("Auto")),
            ("vblank", N_("Buffer swap only")),
            ("off", N_("Off")),
        ]
    },

    "video_sync_method": {
        "default": "",
        "description": N_("Video synchronization method"),
        "type": "choice",
        "values": [
            ("swap", "Swap"),
            ("swap-finish", "Swap-Finish"),
            ("finish-swap-finish", "Finish-Swap-Finish"),
            ("finish-sleep-swap-finish", "Finish-Sleep-Swap-Finish"),
            ("sleep-swap-finish", "Sleep-Swap-Finish"),
            ("swap-fence", "Swap-Fence"),
            ("swap-sleep-fence", "Swap-Sleep-Fence"),
        ]
    },

    "volume": {
        "default": "100",
        "description": N_("Main Volume"),
        "type": "integer",
        "min": 0,
        "max": 100,
    },

    "whdload_splash_delay": {
        "default": "200",
        "description": N_("WHDLoad splash delay"),
        "type": "integer",
        "min": -1,
        "max": 500,
    },

    "window_border": {
        "default": "",
        "description": N_("Show window border and decorations"),
        "type": "boolean",
    },

    "zoom": {
        "default": "auto",
        "description": N_("Zoom Amiga display (crop)"),
        "type": "choice",
        "values": [
            ("auto", N_("Auto")),
            ("auto+border", N_("Auto + Border")),
            ("full", N_("Full Frame")),
            ("640x400", "640x400"),
            ("640x400+border", N_("640x400 + Border")),
            ("640x480", "640x480"),
            ("640x480+border", N_("640x480 + Border")),
            ("640x512", "640x512"),
            ("640x512+border", N_("640x512 + Border")),
            ("704x540", "704x540"),
            ("704x566", "704x566"),
            ("724x566", "724x566"),
        ]
    },

}



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_ARB_point_parameters'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_ARB_point_parameters',error_checker=_errors._error_checker)
GL_POINT_DISTANCE_ATTENUATION_ARB=_C('GL_POINT_DISTANCE_ATTENUATION_ARB',0x8129)
GL_POINT_FADE_THRESHOLD_SIZE_ARB=_C('GL_POINT_FADE_THRESHOLD_SIZE_ARB',0x8128)
GL_POINT_SIZE_MAX_ARB=_C('GL_POINT_SIZE_MAX_ARB',0x8127)
GL_POINT_SIZE_MIN_ARB=_C('GL_POINT_SIZE_MIN_ARB',0x8126)
@_f
@_p.types(None,_cs.GLenum,_cs.GLfloat)
def glPointParameterfARB(pname,param):pass
@_f
@_p.types(None,_cs.GLenum,arrays.GLfloatArray)
def glPointParameterfvARB(pname,params):pass
# Calculate length of params from pname:PointParameterNameARB



import os
from fs_uae_launcher.ui.behaviors.amigaenablebehavior import AmigaEnableBehavior
from fsgs.context import fsgs
import fsui as fsui
from ..Config import Config
from ..CDManager import CDManager
from ..FloppyManager import FloppyManager
from ..I18N import gettext
from .IconButton import IconButton
from .LauncherFilePicker import LauncherFilePicker


class FloppySelector(fsui.Group):

    def __init__(self, parent, drive):
        fsui.Group.__init__(self, parent)
        self.cd_mode = False
        self.drive = drive
        self.config_key = ""
        self.config_key_sha1 = ""

        self.eject_button = IconButton(self, "eject_button.png")
        AmigaEnableBehavior(self.eject_button)
        self.eject_button.set_tooltip(gettext("Eject"))
        self.eject_button.activated.connect(self.on_eject)

        self.text_field = fsui.TextField(self, "", read_only=True)
        AmigaEnableBehavior(self.text_field)

        self.browse_button = IconButton(self, "browse_file_16.png")
        self.browse_button.set_tooltip(gettext("Browse for File"))
        self.browse_button.activated.connect(self.on_browse)
        AmigaEnableBehavior(self.browse_button)

        self.layout = fsui.HorizontalLayout()
        self.layout.add(self.eject_button)
        self.layout.add_spacer(10)
        self.layout.add(self.text_field, expand=True)
        self.layout.add_spacer(10)
        self.layout.add(self.browse_button)

        # Config.add_listener(self)
        #  fsgs.signal.connect(self.on_config,
        #          "fsgs:config:floppy_drive_{0}".format(self.drive),
        #          "fsgs:config:cdrom_drive_{0}".format(self.drive))
        fsgs.signal.connect("config", self.on_config)
        self.update_config_key()

    def on_destroy(self):
        # fsgs.signal.disconnect(
        #     "fsgs:config:floppy_drive_{0}".format(self.drive),
        #     self.on_config_floppy_drive)
        fsgs.signal.disconnect("config", self.on_config)

    def enable(self, enable):
        self.text_field.enable(enable)
        self.browse_button.enable(enable)
        self.eject_button.enable(enable)

    def on_config(self, key, value):
        if key != self.config_key:
            return
        dir, name = os.path.split(value)
        if dir:
            path = "{0} ({1})".format(name, dir)
        else:
            path = name
        self.text_field.set_text(path)
        self.text_field.set_cursor_position(0)

    def update_config_key(self):
        if self.cd_mode:
            self.config_key = "cdrom_drive_{0}".format(self.drive)
            self.config_key_sha1 = "x_cdrom_drive_{0}_sha1".format(self.drive)
        else:
            self.config_key = "floppy_drive_{0}".format(self.drive)
            self.config_key_sha1 = "x_floppy_drive_{0}_sha1".format(self.drive)
        self.on_config(self.config_key, Config.get(self.config_key))

    def set_cd_mode(self, cd_mode):
        self.cd_mode = cd_mode
        self.update_config_key()

    def on_eject(self):
        if self.cd_mode:
            CDManager.eject(self.drive)
        else:
            FloppyManager.eject(self.drive)

    def on_browse(self):
        if self.cd_mode:
            title = gettext("Choose CD-ROM Image")
            # default_dir = FSGSDirectories.get_cdroms_dir()
            type = "cd"
        else:
            title = gettext("Choose Floppy Image")
            # default_dir = FSGSDirectories.get_floppies_dir()
            type = "floppy"
        dialog = LauncherFilePicker(self.get_window(), title, type,
                                    Config.get(self.config_key))

        if not dialog.show_modal():
            return
        path = dialog.get_path()

        if self.cd_mode:
            fsgs.amiga.insert_cd(self.drive, path)
        else:
            fsgs.amiga.insert_floppy(self.drive, path)

        # checksum_tool = ChecksumTool(self.get_window())
        # if self.cd_mode:
        #     sha1 = ""
        #     print("FIXME: not calculating CD checksum just yet")
        # else:
        #     sha1 = checksum_tool.checksum(path)
        # path = Paths.contract_path(path, default_dir)
        # Config.set_multiple([
        #         (self.config_key, path),
        #         (self.config_key_sha1, sha1)])



from fsui import Window, VerticalLayout
from .iconview import IconView
from .vfs import get_vfs_item


class VFSIconView(IconView):

    def __init__(self, parent, item):
        IconView.__init__(self, parent)
        self.item = item
        self.children = sorted(self.item.children())

    def get_item_count(self):
        return len(self.children)

    def get_item_text(self, index):
        return self.children[index]

    def get_item_icon(self, index):
        name = self.children[index]
        item = self.item.item(name)
        icon = item.icon()
        # return fsui.Image("")
        return icon.image()

    def on_activate_item(self, index):
        name = self.children[index]
        item = self.item.item(name)
        # window = VFSWindow(item)
        # window.show()
        VFSWindow.open_item(item)


class VFSWindow(Window):

    opened_count = 0

    @classmethod
    def open_item(cls, item):
        content_type = item.content_type()
        if content_type == "application/x-fs-uae-application":
            item.open()
        elif content_type == "application/x-fs-uae-vfolder":
            uri = item.uri()
            # try:
            #     window = window_uris[uri]
            # except KeyError:
            #     window = VFSWindow(item)
            #     window.show()
            # else:
            #     assert isinstance(window, Window)
            #     window.raise_()
            #    return
            from .shell import raise_window, register_window
            if not raise_window(uri):
                window = VFSWindow(item)
                window.show()
                register_window(uri, window)

    @classmethod
    def open_uri(cls, uri):
        item = get_vfs_item(uri)
        cls.open_item(item)

    def __init__(self, item):
        Window.__init__(self)
        self.setWindowTitle(item.name())
        self.layout = VerticalLayout()
        self.icon_view = VFSIconView(self, item)
        self.icon_view.setStyleSheet("background-color: #ededed")
        self.layout.add(self.icon_view, expand=True, fill=True)
        w, h = 600, 400
        self.set_size((w, h))



import fsui as fsui
from .FloppiesGroup import FloppiesGroup
from .MediaListGroup import MediaListGroup
from .Skin import Skin


class FloppiesPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)

        self.floppies_group = FloppiesGroup(self, 4)
        self.media_list_group = MediaListGroup(self, False)

        self.layout = fsui.VerticalLayout()
        self.layout.add(self.floppies_group, fill=True)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        self.layout.add(self.media_list_group, expand=True, fill=True)



from fsgs.platform import PlatformHandler
from fsgs.mednafen.super_nintendo import SuperNintendoRunner
from .loader import SimpleLoader


class SuperNintendoPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Super Nintendo"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return SuperNintendoLoader(fsgs)

    def get_runner(self, fsgs):
        return SuperNintendoRunner(fsgs)


class SuperNintendoLoader(SimpleLoader):
    pass



import traceback
from fsgs.input.manager import DeviceManager
from fsgs.input.inputdevice import InputDevice
from fsgs.runner import GameRunner
from game_center.glui.opengl import *
from game_center.glui.menu import Menu
from game_center.glui.font import Font
from game_center.glui.render import Render
from game_center.glui.state import State
from game_center.glui.items import GameCenterItem, HomeItem, MenuItem
from game_center.glui.input import InputHandler
from .launchmenu import LaunchMenu


class InputDevices(object):

    device_list_version = None


class DeviceDataDict(dict):

    def __getitem__(self, item):
        return dict.__getitem__(self, item.upper())

    def __setitem__(self, key, value):
        return dict.__setitem__(self, key.upper(), value)


class InputMenu(Menu):

    def __init__(self, item, controller):
        Menu.__init__(self)
        # self.top_menu_transition = 0.0
        self.items.append(item)
        if self.use_game_center_item():
            self.top.left.append(GameCenterItem())
        self.top.left.append(HomeItem())
        self.top.left.append(MenuItem(item.title))

        self.controller = controller
        assert isinstance(self.controller, GameRunner)
        self.first_shown_at = 0

        # controller must now initialize input ports
        # self.controller.init_input()

        # for input in self.controller.inputs:
        #     input["device_id"] = None

        # get all available input devices
        # self.devices, info = InputDevice.get_input_devices(
        #     "", 0, 100, version=2)

        device_manager = InputHandler.get_device_manager()
        # self.devices = DeviceManager.instance().get_devices()
        self.devices = device_manager.get_devices()

        self.device_list_version = InputDevices.device_list_version
        self.device_data = DeviceDataDict()
        # [{"index": 0} for x in self.devices]
        for device in self.devices:
            self.device_data[device.id] = {"port": 0, "device": device}
            self.check_device(self.device_data[device.id])
        # FIXME: Make InputHandler / InputDevice set variables
        # etc and self detect when to reinit?
        # InputHandler.reinit_joysticks()

        # calling activate to try to set the active input device
        # to the first input port, if possible
        # self.activate()
        self.set_defaults()

    def set_defaults(self):
        print("set_defaults")
        devices = []
        for device in self.devices:
            score = 0
            if device.id == InputHandler.last_device:
                score = -1
            devices.append([score, device])
        devices = [x[1] for x in sorted(devices)]
        print("- devices:", devices)
        for i, input in enumerate(self.controller.ports):
            print("- input port {0}:".format(i))
            for device in devices:
                try:
                    device.configure(input.mapping_name)
                except Exception:
                    traceback.print_exc()
                    pass
                else:
                    print("  -> device id", device.id)
                    input.device_id = device.id
                    input.device = device
                    devices.remove(device)
                    self.device_data[device.id]["ok"] = True
                    self.device_data[device.id]["port"] = i
                    break

    def go_left(self, count=1):
        # device_id = InputHandler.last_device
        # for i, device in enumerate(self.devices):
        #     if not device.name == device_name:
        #         continue
        try:
            device_data = self.device_data[InputHandler.last_device]
        except KeyError as e:
            print(repr(e))
            return
        if device_data["port"] > 0:
            device_data["port"] -= 1
            # device_data["port"] -= 1
            self.check_device(device_data)

    def go_right(self, count=1):
        try:
            device_data = self.device_data[InputHandler.last_device]
        except KeyError as e:
            print(InputHandler.last_device)
            print(self.device_data)
            print(repr(e))
            return
        if device_data["port"] < len(self.controller.ports) - 1:
            device_data["port"] += 1
            # device_data["port"] += 1
            self.check_device(device_data)

        # device_name = InputHandler.last_device
        # for i, device in enumerate(self.devices):
        #     if not device.id == device_name:
        #         continue
        #     index = self.device_column[i]
        #     if index < len(self.controller.inputs) - 1:
        #         index += 1
        #         self.device_column[i] = index
        #         self.check_device(device)
        #     break

    def on_status(self, status):
        print("received status", status)

    def run_game(self):
        new_menu = LaunchMenu(self.items[0], self.controller)

        State.history.append(new_menu)
        # FIXME
        from game_center.glui.window import set_current_menu
        set_current_menu(new_menu)

        # print("run_game, controller = ", id(self.controller))
        #
        # # FIXME:
        # from game_center.glui.window import run_game
        # #run_game({"controller": self.controller})
        # run_game(self.controller, self.on_status)
        #
        # go_back_to_game_selection = True
        #
        # # FIXME:
        # from game_center.glui.window import go_back
        # if go_back_to_game_selection:
        #     go_back()
        #     go_back()
        # else:
        #     # recreate controller for game menu
        #     #State.history[-2].recreate_controller()
        #     go_back()

    def activate(self):
        if len(self.controller.ports) == 0:
            return self.run_game()
        # print(self.device_data)
        # print("InputHandler.last_device", InputHandler.last_device)
        try:
            device_data = self.device_data[InputHandler.last_device]
        except KeyError as e:
            print(repr(e))
            return
        # device_data = self.device_data[InputHandler.last_device]
        if not device_data["ok"]:
            return
        port = device_data["port"]
        device = device_data["device"]
        # print("activate", port, device)
        for i in range(len(self.controller.ports)):
            if i == port:
                if i == 0 and device.id == self.controller.ports[i].device_id:
                    return self.run_game()   
                self.controller.ports[i].device_id = device.id
                self.controller.ports[i].device = device
            elif self.controller.ports[i].device_id:
                # remove this device from other ports

                # if self.controller.inputs[i]["device"].id == device.id:
                if self.controller.ports[i].device_id == device.id:
                    self.controller.ports[i].device_id = None
                    self.controller.ports[i].device = None

    def check_device(self, device_data):
        device = device_data["device"]
        port = device_data["port"]
        try:
            # print("configuring device", device, "for",
            # self.controller.ports[port].mapping_name)
            device.configure(self.controller.ports[port].mapping_name)
        except InputDevice.MissingPlatformSupportException:
            device_data["ok"] = False
        except Exception as e:
            print(repr(e))
            device_data["ok"] = False
        else:
            device_data["ok"] = True
        # print(device, index, self.controller.inputs[index]["type"], ok)

    def render(self):
        if self.first_shown_at == 0:
            self.first_shown_at = State.time    
        # FIXME:
        # from .gamemenu import render_wall, render_screen
        # glClear(GL_DEPTH_BUFFER_BIT)
        # render_wall()
        # render_screen()
        
        Render.hd_perspective()
        if len(self.controller.ports) == 0:
            color = (1.0, 0.0, 0.0, 1.0)
            text = "No input configuration needed"
            Render.text(text, Font.main_path_font, 200, 730, 400,
                        color=color, halign=0)
            text = "Press enter or primary key to start game"
            Render.text(text, Font.main_path_font, 200, 670, 400,
                        color=color, halign=0)
            return
        
        if self.device_list_version != InputDevices.device_list_version:
            print(" -- device list version changed")
            # self.devices, info = InputDevice.get_input_devices("", 0, 100,
            # version=2)
            self.devices = DeviceManager.instance().get_devices()
            self.device_list_version = InputDevices.device_list_version
            # [{"index": 0} for x in self.devices]
            device_ids = set()
            for device in self.devices:
                device_ids.add(device.id)
                try:
                    self.device_data[device.id]["device"] = device
                except KeyError:
                    print(" -- add device info for", device.id)
                    self.device_data[device.id] = {
                        "port": 0, "device": device}
                    self.check_device(self.device_data[device.id])
            for data_key in self.device_data.keys():
                if not data_key in device_ids:
                    print(" -- removing device_data for", data_key)
                    del self.device_data[data_key]
            for input in self.controller.ports:
                if not input["device_id"] in device_ids:
                    print(" -- removing device from input", input.device_id)
                    input["device_id"] = None
        
        for port, input in enumerate(self.controller.ports):
            center_x = 400 + 400 * port
            
            text = input.name.upper()
            color = (1.0, 0.0, 0.0, 1.0)
            Render.text(text, Font.main_path_font, center_x - 200, 760, 400,
                        color=color, halign=0)
            text = input.description.upper()
            color = (1.0, 0.0, 0.0, 1.0)
            Render.text(text, Font.main_path_font, center_x - 200, 730, 400,
                        color=color, halign=0)
            if input.device_id:
                device = self.device_data[input.device_id]["device"]
                color = (1.0, 0.5, 0.5, 1.0)
                text = device.name.upper()
                Render.text(text, Font.main_path_font, center_x - 200, 680,
                            400, color=color, halign=0)

            for j, device in enumerate(self.devices):
                device_data = self.device_data[device.id]
                # print(1, repr(device))
                if device_data["port"] != port:
                    continue
                text = device.name.upper()
                if device_data["ok"]:
                    color = (1.0, 1.0, 1.0, 1.0)
                else:
                    color = (0.5, 0.5, 0.5, 1.0)
                Render.text(text, Font.main_path_font, center_x - 200,
                            600 - j * 40, 400, color=color, halign=0)
        
        fade = 1.0 - (State.time - self.first_shown_at) * 3.0
        if fade > 0.0:
            Render.dirty = True
            fs_emu_blending(True)
            fs_emu_texturing(False)
            glDisable(GL_DEPTH_TEST)
            Render.hd_perspective()
            glBegin(GL_QUADS)
            glColor4f(0.0, 0.0, 0.0, fade)
            glVertex2f(0, 0)
            glVertex2f(1920, 0)
            glVertex2f(1920, 1080)
            glVertex2f(0, 1080)            
            glEnd()
            glEnable(GL_DEPTH_TEST)



import fsui
from fs_uae_workspace.shell import SimpleApplication
from fs_uae_launcher.ui.settings.VideoSettingsPage import VideoSettingsPage
from fs_uae_launcher.res import gettext


class SettingsWindow(fsui.Window):

    def __init__(self):
        fsui.Window.__init__(self, None, gettext("Video Settings"))
        self.layout = fsui.VerticalLayout()
        page = VideoSettingsPage(self)
        page.set_min_width(700)
        self.layout.add(page, fill=True)
        self.set_size(self.layout.get_min_size())

    def __del__(self):
        print("SettingsWindow.__del__")


application = SimpleApplication(SettingsWindow)



"""
The goals of this class are:
1. To be able to use ISO-8859-1 encoding instead of CP43, and
2. Also support zip files with UTF-8 encoding, and

Python 3 assumes strings must be converted to CP437 for non-ASCII characters.
This is why there'll be decoding/encoding from/to CP437 even for ISO-8859-1
and UTF-8 name lookup (with Python 3). Complications arise because Python
seems to detect UTF-8 in some zip files (based on a flag) and then
automatically convert to/from unicode.
"""
import zipfile
from typing import List


class ZipFile(zipfile.ZipFile):

    def __init__(self, path: str, mode: str="r") -> None:
        zipfile.ZipFile.__init__(self, path, mode)

    def getinfo(self, name: str) -> zipfile.ZipInfo:
        try:
            n = name.encode("ISO-8859-1").decode("CP437")
            return zipfile.ZipFile.getinfo(self, n)
        except Exception:
            pass
        try:
            n = name.encode("UTF-8").decode("CP437")
            return zipfile.ZipFile.getinfo(self, n)
        except Exception:
            pass
        return zipfile.ZipFile.getinfo(self, name)

    def namelist(self) -> List[str]:
        names = zipfile.ZipFile.namelist(self)
        for i, name in enumerate(names):
            print("...", repr(name))
            try:
                name = name.encode("CP437").decode("UTF-8")
            except Exception:
                try:
                    name = name.encode("CP437").decode("ISO-8859-1")
                except Exception:
                    pass
            names[i] = name
        return names



#!/usr/bin/env python3

import os
import sys
import shutil
import subprocess


# def fix_binary(path, frameworks_dir):
def fix_binary(path, macos_dir):
    args = ["file", path] 
    p = subprocess.Popen(args, stdout=subprocess.PIPE)
    data = p.stdout.read().decode("UTF-8")
    p.wait()
    if "Mach-O" not in data:
        return 0

    print("fixing", path)
    changes = 0
    if not os.path.exists(path):
        raise Exception("could not find " + repr(path))
    args = ["otool", "-L", path] 
    p = subprocess.Popen(args, stdout=subprocess.PIPE)
    data = p.stdout.read().decode("UTF-8")
    p.wait()
    for line in data.split('\n'):
        line = line.strip()
        if not line:
            continue
        if line.startswith("/usr/lib") or line.startswith("/System"):
            # old = line.split(' ')[0]
            # print("ignoring", old)
            continue
        if line.startswith("@executable_path"):
            continue

        old = line.split(" ")[0]
        if old == "@rpath/XCTest.framework/Versions/A/XCTest":
            continue
        if "Contents" in old:
            continue
        print(old)

        if os.path.basename(path) == os.path.basename(old):
            if "/" not in old:
                continue
            os.chmod(path, 0o755)
            args = ["install_name_tool", "-id", os.path.basename(path), path]
            print(args)
            p = subprocess.Popen(args)
            assert p.wait() == 0
            changes += 1
            continue

        # if not os.path.isabs(old):
        #     old = os.path.join(
        #         os.environ["DYLD_FALLBACK_LIBRARY_PATH"].split(":")[0], old)

        old_dir, name = os.path.split(old)
        # new = old.replace(old, '@executable_path/../Frameworks/' + name)
        new = old.replace(old, '@executable_path/' + name)
        # dst = os.path.join(frameworks_dir, os.path.basename(old))
        dst = os.path.join(macos_dir, os.path.basename(old))
        if not os.path.exists(dst):
            print("copying", old)
            shutil.copy(old, dst)
            os.chmod(dst, 0o755)
            changes += 1
        # print(os.path.basename(path), "vs", os.path.basename(old))
        # if os.path.basename(path) == os.path.basename(old):
        #     args = ["install_name_tool", "-id", new, path]
        # else:
        os.chmod(path, 0o755)
        args = ["install_name_tool", "-change", old, new, path]
        print(args)
        p = subprocess.Popen(args)
        assert p.wait() == 0

    return changes


def fix_iteration(app):
    binaries = []
    macos_dir = os.path.join(app, "Contents", "MacOS")
    frameworks_dir = os.path.join(app, "Contents", "Frameworks")
    extra_paths = []
    for dir_path, dir_names, file_names in os.walk(macos_dir):
        for name in file_names:
            p = os.path.join(dir_path, name)
            if os.path.isdir(p):
                pass
            elif p.endswith("_debug.dylib"):
                pass
            else:
                binaries.append(p)
    # for name in os.listdir(macos_dir):
    #     p = os.path.join(macos_dir, name)
    #    if os.path.isdir(p):
    #         extra_paths.append(p)
    #     else:
    #         binaries.append(os.path.join(macos_dir, name))
    # for extra_dir in extra_paths:
    #     for name in os.listdir(extra_dir):
    #         p = os.path.join(extra_dir, name) 
    #         if os.path.isdir(p):
    #             for name2 in os.listdir(p):
    #                 if not name2.endswith("_debug.dylib"):
    #                     binaries.append(os.path.join(p, name2))
    #         else:
    #             binaries.append(p)

    if os.path.exists(frameworks_dir):
        for name in os.listdir(frameworks_dir):
            binaries.append(os.path.join(frameworks_dir, name))
    changes = 0
    for binary in binaries:
        changes += fix_binary(binary, macos_dir)
    return changes


def main():
    app = sys.argv[1]
    while True:
        changes = fix_iteration(app)
        if changes == 0:
            break


if __name__ == "__main__":
    main()



# from fsbc.Application import app
from .client import OGDClient


class SynchronizerContext(object):

    meta = {}

    def __init__(self):
        pass

    @property
    def username(self):
        return OGDClient.get_credentials()[0]

    @property
    def password(self):
        return OGDClient.get_credentials()[1]
